#ifndef __REFLECTANCEMAP_H
#define __REFLECTANCEMAP_H
#include "xp_threads.h"
#include "image.h"
#include "floatimage.h"
#include "DCT.h"


// JPEG block size
#define BLOCK_SIZEX 8
#define BLOCK_SIZEY 8


// do not make this more than 32, or our bit field will overrun
#define MAX_THREADS 32


extern int DCT_initialized;
extern float DCT_bases[BLOCK_SIZEY][BLOCK_SIZEX];
extern float IDCT_bases[BLOCK_SIZEY][BLOCK_SIZEX];
extern unsigned char DCT_temp2[BLOCK_SIZEY][BLOCK_SIZEX];
extern float DCT_threshold[BLOCK_SIZEY][BLOCK_SIZEX];
extern int zigzag[BLOCK_SIZEX * BLOCK_SIZEY];
extern void init_DCT();


class LDR_ReflectanceMapArray;



class LDR_ReflectanceMapArray
{
	public:	
		int max_theta, max_phi;		// number of phi and theta samples per reflectance map
		int sizex, sizey;			// number of reflectance maps stored (2 dimensional array)
		int RM_size;
		int DC_count;				// = RM_size / (BLOCK_SIZEX * BLOCK_SIZEY),  number of DC components per RM
		int chan_size;
		unsigned char *line;
		int RLEbytes;

		char *RLEbuf;
		char **RLEindex;
		unsigned char *RMtemp;



		LDR_ReflectanceMapArray()
		{
			max_theta = max_phi = sizex = sizey = RM_size = DC_count = chan_size = 0;
			line = NULL;
			RLEbuf = NULL;
			RLEindex = NULL;
			RMtemp = NULL;
		}


		void Release()
		{
			if (line)
			{
				delete line;
				line = NULL;
			}
			if (RLEbuf)
			{
				delete RLEbuf;
				RLEbuf = NULL;
			}
			if (RLEindex)
			{
				delete RLEindex;
				RLEindex = NULL;
			}
			if (RMtemp)
			{
				delete RMtemp;
				RMtemp = NULL;
			}
			max_theta = 0;
			max_phi = 0;
			sizex = 0;
			sizey = 0;
			RM_size = 0;
			DC_count = 0;
			chan_size = 0;
		}

		int Allocate(int max_theta, int max_phi, int sizex, int sizey)
		{
			if ((this->max_theta == max_theta) &&
				(this->max_phi == max_phi) &&
				(this->sizex == sizex) &&
				(this->sizey == sizey))
				return 1;								// correct size already

			Release();

			line = new unsigned char[max_theta * max_phi * sizex * sizey * 3];

			if (!line)
				return 0;

			this->max_theta = max_theta;
			this->max_phi = max_phi;
			this->sizex = sizex;
			this->sizey = sizey;
			this->chan_size = max_phi * max_theta;
			this->RM_size = chan_size * 3;
			this->DC_count = chan_size / (BLOCK_SIZEX * BLOCK_SIZEY);
			return 1;
		}

		~LDR_ReflectanceMapArray()
		{
			Release();
		}

		unsigned char *redRM(int x, int y)
		{
			return &line[(y * sizex + x) * RM_size];
		}

		unsigned char *greenRM(int x, int y)
		{
			return &line[(y * sizex + x) * RM_size + chan_size];
		}

		unsigned char *blueRM(int x, int y)
		{
			return &line[(y * sizex + x) * RM_size + chan_size * 2];
		}
		
		void enterImage(image &img, int theta, int phi)
		{
			if ((img.sizex == sizex) && (img.sizey == sizey))
			{
				RGB *src = img.line[0];
				RGB *last = &img.line[sizey-1][sizex-1];
				unsigned char *dst = &line[phi * max_theta + theta];
				while (src <= last)
				{
					*dst = src->rgbRed;
					dst += chan_size;
					*dst = src->rgbGreen;
					dst += chan_size;
					*dst = src->rgbBlue;
					dst += chan_size;
					src++;
				}
			}
			else
			{
				unsigned char *dst = &line[phi * max_theta + theta];
				unsigned char *last = &dst[RM_size * sizex * sizey - 1];
				while (dst <= last)
				{
					*dst = 0;
					dst += chan_size;
					*dst = 0;
					dst += chan_size;
					*dst = 0;
					dst += chan_size;
				}
			}
		}


		void getImage(image &img, int theta, int phi)
		{
			unsigned char *src = &line[phi * max_theta + theta];
			img.Allocate(sizex, sizey);

			RGB *dst = img.line[0];
			RGB *last = &img.line[sizey-1][sizex-1];

			while (dst <= last)
			{
				dst->rgbRed = *src;
				src += chan_size;
				dst->rgbGreen = *src;
				src += chan_size;
				dst->rgbBlue = *src;
				src += chan_size;
				dst++;
			}
		}


		void getRM(image &img, int x, int y)
		{
			unsigned char *r = redRM(x, y);
			unsigned char *g = greenRM(x, y);
			unsigned char *b = blueRM(x, y);

			img.Allocate(max_theta, max_phi);

			for (int i = 0; i < max_phi; i++)
				for (int a = 0; a < max_theta; a++)
				{
					img.line[i][a].rgbBlue = *b++;
					img.line[i][a].rgbGreen = *g++;
					img.line[i][a].rgbRed = *r++;
				}
		}


		void IDCT()
		{
			for (int y = 0; y < sizey; y++)
				for (int x = 0; x < sizex; x++)
				{
					IDCT_RM(redRM(x, y));
					IDCT_RM(greenRM(x, y));
					IDCT_RM(blueRM(x, y));
				}
		}

		void DCT(float quality, float *progress = NULL)
		{
			int count = 0;
			int max = sizex * sizey;
			for (int y = 0; y < sizey; y++)
			{
				for (int x = 0; x < sizex; x++)
				{
					DCT_RM(redRM(x, y), quality);
					DCT_RM(greenRM(x, y), quality);
					DCT_RM(blueRM(x, y), quality);

					if (progress)
						*progress = count / (float) max;
					count++;
				}
			}
		}

		void DCT_RM(unsigned char *rm, float quality)
		{
			unsigned int u, v;
			unsigned int maxu = max_theta / BLOCK_SIZEX;
			unsigned int maxv = max_phi / BLOCK_SIZEY;

			init_DCT();

			for (v = 0; v < maxv; v++) 
			{
				for (u = 0; u < maxu; u++) 
				{
					DCT_Block(rm, u * BLOCK_SIZEX, v * BLOCK_SIZEY, quality);
				}
			}			
		}

		void IDCT_RM(unsigned char *rm)
		{
			unsigned int u, v;
			unsigned int maxu = max_theta / BLOCK_SIZEX;
			unsigned int maxv = max_phi / BLOCK_SIZEY;

			init_DCT();

			for (v = 0; v < maxv; v++) 
			{
				for (u = 0; u < maxu; u++) 
				{
					IDCT_Block((char *) rm, u * BLOCK_SIZEX, v * BLOCK_SIZEY);
				}
			}			
		}

		void DCT_Block(unsigned char *rm, int u0, int v0, float quality);
		void IDCT_Block(char *rm, int u0, int v0);

		void ZigZag_RM(unsigned char *source);		// apply zig-zag transformation to the reflectance map starting at source
		void UnZigZag_RM(unsigned char *source);
		int RLE_compress_RM(unsigned char *source, char *ACdest);			// returns number of bytes written to dest
		float RLE_compress(float *progress = NULL, image *img = NULL);

		void CalculateEnvironment(floatimage &environment);
		int StartCalculation(floatimage &environment, int cur = 0, int max = 1);
		unsigned char computePixelChannel(char *rle, int *mult, int count);
		void calculateLines(int start, int end);

		int SaveMonsta(const char *name);
		int LoadMonsta(const char *name);

		int un_RLE_compress()
		{
			for (int y = 0; y < sizey; y++)
			{
				for (int x = 0; x < sizex; x++)
				{

				}
			}
		}

		char *redRLE(int x, int y)
		{
			return RLEindex[0 * sizex * sizey + y * sizex + x];
		}

		char *greenRLE(int x, int y)
		{
			return RLEindex[1 * sizex * sizey + y * sizex + x];
		}

		char *blueRLE(int x, int y)
		{
			return RLEindex[2 * sizex * sizey + y * sizex + x];
		}
};


#endif
