#ifndef __RENDER_H
#define __RENDER_H

#include "ReflectanceMap.h"
#include "gamma.h"



THREAD_PROTOTYPE(calc_img_thread,lpParameter);


class MonstaRenderer;


typedef struct
{
	MonstaRenderer *rma;
	int draw_strip;
	int max_strips;

	int thread_number;
} thread_params;



class MonstaRenderer
{
	public:
		// Threading stuff
		mutex_t thread_state_mutex;
		cond_t start_calculation_signal;   // signal to start the threads calculating
		unsigned int thread_state;	   // bit n = 0 (thread n is waiting)   bit n = 1 (thread n is not waiting)
		int thread_count;
		int thread_die_flag;
		thread_t thread[MAX_THREADS];
		thread_params tparams[MAX_THREADS];



		image *finished_img;
		floatimage env_dct;
		int left, right;


		int *env_dct_int[3];
		
		int exposure;

		LDR_ReflectanceMapArray *data;

		BYTE linear_to_gamma_table[4096 * 3];



		MonstaRenderer()
		{
			left = right = 0;
			thread_count = 0;
			data = new LDR_ReflectanceMapArray;
			finished_img = new image;
			env_dct_int[0] = NULL;
			env_dct_int[1] = NULL;
			env_dct_int[2] = NULL;
			exposure = 6;

			compute_linear_to_gamma_with_shoulder(linear_to_gamma_table, 0.7);
		}

		~MonstaRenderer()
		{
		  if (thread_count)		 // kill off the threads
			{
				for (int x = 0; x < thread_count; x++)
					thread_kill(&thread[x]);
				thread_count = 0;
				}

			if (data)
				delete data;
			if (finished_img)
				delete finished_img;
			if (env_dct_int[0])
			{
				delete env_dct_int[0];
				delete env_dct_int[1];
				delete env_dct_int[2];
			}
		}

		void LoadMonsta(const char *filename)
		{
			data->LoadMonsta(filename);
			finished_img->Allocate(data->sizex, data->sizey);

			env_dct_int[0] = new int [data->max_phi * data->max_theta];
			env_dct_int[1] = new int [data->max_phi * data->max_theta];
			env_dct_int[2] = new int [data->max_phi * data->max_theta];

			left = 0;
			right = data->sizex - 1;
		}

		int CalculationFinished()
		{
			return !thread_state;
		}


		void InitThreads(int num_threads);						// At most 32 threads please

		void CalculateEnvironment(floatimage &environment);
		int StartCalculation(floatimage &environment, int cur = 0, int max = 1);
		unsigned char computePixelChannel(char *rle, int *mult, int count);
		void calculateLines(int start, int end);

		image &GetImage()
		{
			return *finished_img;
		}
};


class MipMapMonstaRenderer : public MonstaRenderer
{
	public:
		LDR_ReflectanceMapArray *datalevel[3];
		image imglevel[3];

		int mode;
		int calclevel;
		int displevel;

		MipMapMonstaRenderer()
		{
		  for(int i = 0; i < 3; i++)
		    datalevel[i] = NULL;
		  
		  left = right = 0;
		  thread_count = 0;
		  data = datalevel[0];
		  finished_img = imglevel;
		  env_dct_int[0] = NULL;
		  env_dct_int[1] = NULL;
		  env_dct_int[2] = NULL;
		  exposure = 6;
		  compute_linear_to_gamma_with_shoulder(linear_to_gamma_table, 0.7);
			
		  mode = 0;
		  calclevel = 0;
		  displevel = 0;
		}

		~MipMapMonstaRenderer()
		{
			if (thread_count)		// kill off the threads
			{
				for (int x = 0; x < thread_count; x++)
					thread_kill(&thread[x]);
				thread_count = 0;
			}

			if (env_dct_int[0])
			{
				delete env_dct_int[0];
				delete env_dct_int[1];
				delete env_dct_int[2];
			}


#if 0 // ?
			for(int i = 0; i < 3; i++)
			  if(datalevel[i])
			    delete datalevel[i];
#endif

		}

		void LoadMipMapLevel(int lev, const char *filename)
		{
			if ((lev < 0) || (lev >= 3))
				return;

			LDR_ReflectanceMapArray *temp;

			temp = new LDR_ReflectanceMapArray;

			temp->LoadMonsta(filename);

			datalevel[lev] = temp;

			imglevel[lev].Allocate(datalevel[lev]->sizex, datalevel[lev]->sizey);

			if (!env_dct_int[lev])
			  env_dct_int[lev] = new int [datalevel[lev]->max_phi * datalevel[lev]->max_theta];
	     
		}

		void SetMipMapLevel(int lev, LDR_ReflectanceMapArray *monsta)
		  {
		    if ((lev < 0) || (lev >= 3))
		      return;

		    datalevel[lev] = monsta;

		    imglevel[lev].Allocate(datalevel[lev]->sizex, datalevel[lev]->sizey);

		    if (!env_dct_int[lev])
		      env_dct_int[lev] = new int [datalevel[lev]->max_phi * datalevel[lev]->max_theta];
		      
		  }

		void NewCalculation(floatimage &environment)
		{
			mode = 0;
			calclevel = 0;
			displevel = 0;

			data = datalevel[calclevel];
			finished_img = &imglevel[calclevel];

			StartCalculation(environment, 0, 1);
		}

		void Refine(floatimage &environment)
		{
			if ((mode >= 20) || (mode < 0))
				return;			// no refinement necessary, or error
			
			mode++;

			int cur;
			int max;

			if (mode <= 4)
			{
				cur = mode - 1;
				max = 4;
				calclevel = 1;
				if (mode == 4)
					displevel = calclevel;
			}
			else
			{
				cur = mode - 5;
				max = 16;
				calclevel = 2;
				if (mode == 20)
					displevel = calclevel;
			}
			
			data = datalevel[calclevel];
			finished_img = &imglevel[calclevel];

			StartCalculation(environment, cur, max);
		}

		image* GetImage()
		{
			return &(imglevel[displevel]);
		}
};


#endif
