#include <math.h>
#include "floatimage.h"
#include "rad_color.h"
#include "rad_header.h"
#include "rad_resolu.h"
#include "stdio.h"
#include "memory.h"
#include <malloc.h>

#include "bmp.h"
#include "gamma.h"
#include "DCT.h"

// floating point TIFF format headers
#ifdef FLOATIMAGE_USETIFF
  #include "tiffio.h"
#endif

static inline void REVERSEBYTES(float *source, float *dest)
{
  char *j = (char*) source;
  char *dj = (char*) dest;
  dj[0] = j[3];
  dj[1] = j[2];
  dj[2] = j[1];
  dj[3] = j[0];
}

static inline void REVERSEBYTES_INPLACE(float *source)
{
  char *j = (char*) source;
  j[0] ^= j[3];
  j[3] ^= j[0];
  j[0] ^= j[3];
  j[1] ^= j[2];
  j[2] ^= j[1];
  j[1] ^= j[2];
}


#define  subcolor(c1,c2)	((c1)[0]-=(c2)[0],(c1)[1]-=(c2)[1],(c1)[2]-=(c2)[2])
#define  lumcolor(c1)	(0.3f*(c1)[0]+0.6f*(c1)[1]+0.1f*(c1)[2])


#ifndef M_PI
# define M_PI	3.14159265358979323846264338327950288
#endif

#define m_boxvariance(n) (((2*(n)*(n)*(n)+3*(n)*(n)+(n))/3.f)*(4*(n)+2)/((2*(n)+1)*(2.f*(n)+1)))



// Floating point format:
//
//			SEEEEEEEEMMMMMMMMMMMMMMMMMMMMMMM
//			|       |       |       |      |
//			0       8       16      24     31
//
//
//		value = (-1)^(S) * (1.MMMMMMMMMMMMMMMMMMMMMMM) * 2^(signed binary EEEEEEEE)
//

inline unsigned short int float2short(float f)
{
  unsigned int bits = * (unsigned int *) &f;
  return (bits >> 15);
}


inline unsigned short int signedfloat2short(float f)
{
  unsigned int bits = * (unsigned int *) &f;
  return (bits >> 16);
}


inline float short2float(unsigned short int s)
{
	unsigned int bits = s << 15;
	return * (float *) &bits;
}

inline float short2signedfloat(unsigned short int s)
{
	unsigned int bits = s << 16;
	return * (float *) &bits;
}


void floatimage::ClipLower(float min, float replace)
{
  SaveUndo();

	float *f = &line[0][0].r;
	float *last = &line[sizey-1][sizex-1].b;

	while (f <= last)
	{
		float r = f[0];
		float g = f[1];
		float b = f[2];

		if ((f[0] < min) && (f[1] < min) && (f[2] < min))
		{
			f[0] = replace;
			f[1] = replace;
			f[2] = replace;
		}

		f += 3;
	}
	NotifyObservers();
}



floatimage&	floatimage::operator = (const floatimage &in) 
{
  SaveUndo();

	float *f = &in.line[0][0].r;
	float *last = &in.line[in.sizey-1][in.sizex-1].b;
	float *fout;

	Allocate(in.sizex,in.sizey);
	fout = &line[0][0].r;
	
	while (f<=last)
	{
		*fout++ = *f++;
		*fout++ = *f++;
		*fout++ = *f++;
	}

	NotifyObservers();
	return *this;
}


floatimage::floatimage(const floatimage &in) 
{
  SaveUndo();

	float *f = &in.line[0][0].r;
	float *last = &in.line[in.sizey-1][in.sizex-1].b;
	float *fout;

	Allocate(in.sizex,in.sizey);
	fout = &line[0][0].r;
	
	while (f<=last)
	{
		*fout++ = *f++;
		*fout++ = *f++;
		*fout++ = *f++;
	}

	NotifyObservers();
}



void floatimage::Allocate(int x, int y)
{
	if (line)
	{
		if ((sizex == x) && (sizey == y))
			return;	     // already a good size!  Yay!
		ReleaseImage();
	}

	line = (fRGB **) malloc(sizeof(fRGB *) * y);

	if (line)
	{
		line[0] = (fRGB *) malloc(sizeof(fRGB) * x * y);

		if (!line[0])
			ReleaseImage();
		else
			for (int cy = 1; cy < y; cy++)
				line[cy] = &line[0][cy * x];

		sizex = x;
		sizey = y;
	}
}



void floatimage::ReleaseImage()
{
	if (line)
	{
		if (line[0])
		{
			free(line[0]);
			line[0] = NULL;
		}
		free(line);
		line = NULL;
	}
	sizex = 0;
	sizey = 0;
}

#ifdef _WIN32
#define strcasecmp _stricmp  // we love you, Bill! (though strcasecmp is admittedly a BSD-ism)
#endif

int floatimage::LoadByExtension(char *filename)
{
  if(!filename)
    return 0;

  // make sure we can open the file
  FILE *f = fopen(filename, "rb");
  if(!f)
    return 0;
  fclose(f);

  // find the extension
  char *ext = strrchr(filename, (int) '.');

  if(!ext)
    return 0;

  if(!strcasecmp(ext,".rgbe"))
    return LoadHDR(filename);

  else if(!strcasecmp(ext,".hdr"))
    return LoadHDR(filename);

  else if(!strcasecmp(ext,".pic"))
    return LoadHDR(filename);

  else if(!strcasecmp(ext,".pfm"))
    return LoadPFM(filename);

#ifdef FLOATIMAGE_USETIFF
  else if(!strcasecmp(ext,".tif"))
    return LoadTIF(filename);
#endif

  else
    return 0; // unrecognized extension
}

int floatimage::SaveByExtension(char *filename)
{
  if(!filename)
    return 0;

  // find the extension
  char *ext = strrchr(filename, (int) '.');

  if(!ext)
    return 0;

  if(!strcasecmp(ext,".rgbe"))
    return SaveHDR(filename);
  
  else if(!strcasecmp(ext,".hdr"))
    return SaveHDR(filename);

  else if(!strcasecmp(ext,".pic"))
    return SaveHDR(filename);
  
  else if(!strcasecmp(ext,".pfm"))
    return SavePFM(filename);

#ifdef FLOATIMAGE_USETIFF
  else if(!strcasecmp(ext,".tif"))
    return SaveTIF(filename);
#endif

  else
    return 0; // unrecognized extension
}


int floatimage::LoadHDR(const char *filename)
{
  SaveUndo();

    FILE *infile = fopen(filename, "rb");
	int new_sizex,new_sizey;

	if (infile == NULL)
		return 0;

	getheader(infile);

	int mode = fgetresolu( &new_sizex, &new_sizey, infile);

	Allocate(new_sizex, new_sizey);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	for (int y = 0; y < sizey; y++)
	{
		if (freadscan((COLOR *) line[y], sizex, infile) < 0) 
		{ 
			ReleaseImage();
			fclose(infile);
			return 0;
		}
	}

	fclose(infile);

	if (mode != (YMAJOR | YDECR))
		return 2;

	if(name)
	  free(name);

	name = strdup(filename);

	NotifyObservers();
	changed = 0;

	return 1;
}

int floatimage::SaveHDR(const char *filename)
{
	FILE *outfile = fopen(filename, "wb");

    fprintf(outfile, "#?RADIANCE");
	fputc(10, outfile);
    fprintf(outfile, "# %s", "Made with 100% pure HDR Shop");
	fputc(10, outfile);
    fprintf(outfile, "FORMAT=32-bit_rle_rgbe");
	fputc(10, outfile);
    fprintf(outfile, "EXPOSURE=%25.13f", 1.0);
	fputc(10, outfile);
	fputc(10, outfile);
    fprintf(outfile, "-Y %d +X %d", sizey, sizex);
	fputc(10, outfile);
	
	for (int y = 0; y < sizey; y++)
	{
		if (fwritescan((COLOR *) &line[y][0].r, sizex, outfile) < 0)
		{
			// error
			fclose(outfile);
			return 0;
		}
	}

	fclose(outfile);

	if(name)
	  free(name);

	name = strdup(filename);
	NotifyObservers();
	changed = 0;

	return 1;
}

int floatimage::LoadPFM(const char *filename)
{
  SaveUndo();

	FILE *infile = fopen(filename, "rb");

	int a = fgetc(infile);
	int b = fgetc(infile);
	int junk = fgetc(infile);

	if ((a != 'P') || ((b != 'F') && (b != 'f')))
	{
		fclose(infile);
//		MessageBox(NULL, "Invalid PFN header", blah->name, MB_OK);
		return 0;
	}

	b = (b == 'F');

	fscanf(infile, "%d %d%c", &sizex, &sizey, &junk);
	if ((sizex <= 0) || (sizey <= 0))
	{
		fclose(infile);
//		MessageBox(NULL, "Invalid PFN image size", blah->name, MB_OK);
		return 0;
	}

	float scalefactor;
	fscanf(infile, "%f%c", &scalefactor, &junk);


	Allocate(sizex, sizey);
	if (!line)
	{
		fclose(infile);
//		MessageBox(NULL, "Cannot Allocate Enough Memory", "HDRView", MB_OK);
		return 0;
	}

	a = sizex * (b ? 3 : 1);
	float *fbuf = new float[a];
	fRGB *cur = &line[0][0];
	for (unsigned int y = 0; y < sizey; y++)
	{
		if (fread(fbuf, sizeof(float), a, infile) != a)
		{
			fclose(infile);
			delete fbuf;
//			MessageBox(NULL, "Error Reading Image File", blah->name, MB_OK);
			return 0;
		}

		float *temp = fbuf;
		for (unsigned int x = 0; x < sizex; x++)
		{
			if (b)			// color
			{
				if (scalefactor > 0.0)		// MSB
				{
					REVERSEBYTES(temp++, &cur->r);
					REVERSEBYTES(temp++, &cur->g);
					REVERSEBYTES(temp++, &cur->b);
				}
				else						// LSB
				{
					cur->r = *temp++;
					cur->g = *temp++;
					cur->b = *temp++;
				}
			}
			else			// black and white
			{
				float c;
				if (scalefactor > 0.0)		// MSB
				{
					REVERSEBYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = c;
			}
			cur++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical(0);
	
	if(name)
	  free(name);

	name = strdup(filename);
	
	NotifyObservers();
	changed = 0;

	return 1;
}


int floatimage::SavePFM(const char *filename)
{
	FILE *outfile = fopen(filename, "wb");

	if(!outfile) {
	  printf("Can't open %s for writing\n", filename);
	  return 0;
	}


	fputc('P', outfile);
	fputc('F', outfile);
	fputc(0x0a, outfile);

	fprintf(outfile, "%d %d", sizex, sizey);
	fputc(0x0a, outfile);

	fprintf(outfile, "%f", -1.0f);
	fputc(0x0a, outfile);

	for (int y = sizey - 1; y >= 0; y--)
		fwrite(&line[y][0].r, sizeof(float), 3 * sizex, outfile);

	fclose(outfile);

	// FlipVertical();
	if(name)
	  free(name);

	name = strdup(filename);
	changed = 0;

	return 1;
}

int floatimage::SaveRAW(const char *filename)
{
	FILE *outfile = fopen(filename, "wb");

	if(!outfile) {
	  printf("Can't open %s for writing\n", filename);
	  return 0;
	}

	for(unsigned int i = 0; i < sizey; i++)
	  fwrite(&line[i][0].r, sizeof(float), 3 * sizex, outfile);
	

	fclose(outfile);

	if(name)
	  free(name);

	name = strdup(filename);
	changed = 0;

	return 1;
}

int floatimage::LoadRAW(const char *filename, int width, int height, int rgb, int byte_reverse, int header_bytes = 0)
{
  SaveUndo();

	FILE *infile = fopen(filename, "rb");
	if (!infile)
	{
		return 0;
	}
	
	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	if(header_bytes > 0)
	  {
	    while(header_bytes--)
	      fgetc(infile);
	  }

	int numfloats = (rgb ? 3 * width : width);
	float *fbuf = new float[numfloats];

	fRGB *cur = &line[0][0];
	for (int y = 0; y < height; y++)
	{
		if (fread(fbuf, sizeof(float), numfloats, infile) != numfloats)
		{
			delete fbuf;
			fclose(infile);
//			MessageBox(NULL, "Error Reading Image File", blah->name, MB_OK);
			return 0;
		}

		float *temp = fbuf;
		for (int x = 0; x < width; x++)
		{
			if (rgb)
			{
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &cur->r);
					REVERSEBYTES(temp++, &cur->g);
					REVERSEBYTES(temp++, &cur->b);
				}
				else						// LSB
				{
					cur->r = *temp++;
					cur->g = *temp++;
					cur->b = *temp++;
				}
			}
			else
			{
				float c;
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = c;
			}

			cur ++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();

	if(name)
	  free(name);
	name = strdup(filename);
       
	NotifyObservers();
	changed = 0;
	
	return 1;
}

#ifdef FLOATIMAGE_USETIFF
int floatimage::LoadTIF(const char *filename)
{
  SaveUndo();

	long ImageWidth = 0;
	long ImageLength = 0;
	long BitsPerSample = 0;
	long SamplesPerPixel = 0;

	int CHANNELS;
	int PIXELS;

	float **data;
	void *scanline;
	int i, j;
	TIFF *tiffImage;

	tiffImage = TIFFOpen(filename, "ru");

	if (!tiffImage) 
		return 0;

	TIFFGetField(tiffImage, TIFFTAG_IMAGEWIDTH, &ImageWidth);
	TIFFGetField(tiffImage, TIFFTAG_IMAGELENGTH, &ImageLength);
	TIFFGetField(tiffImage, TIFFTAG_SAMPLESPERPIXEL, &SamplesPerPixel);
	TIFFGetField(tiffImage, TIFFTAG_BITSPERSAMPLE, &BitsPerSample);

	CHANNELS = SamplesPerPixel;
	PIXELS = ImageWidth * ImageLength;

//	if ((CHANNELS != 1) && (CHANNELS != 3))
//		return 0;

	Allocate(ImageWidth, ImageLength);

	if (!line)
		return 0;

	scanline = _TIFFmalloc(TIFFScanlineSize(tiffImage));
	_TIFFmemset(scanline, 0, TIFFScanlineSize(tiffImage));
    
	for (i = 0; i < sizey; i++)
	{
		TIFFReadScanline(tiffImage, scanline, i, 0);

		if (CHANNELS==1)
		{
			for (j = 0; j < sizex; j++)
			{
				float c = ((float *) scanline)[j];
				line[i][j].Set(c, c, c);
			}
		}
		else
		{
			for (j = 0; j < sizex; j++)
			{
				float r, g, b;
				r = ((float *) scanline)[j * CHANNELS];
				g = ((float *) scanline)[j * CHANNELS + 1];
				b = ((float *) scanline)[j * CHANNELS + 2];
				line[i][j].Set(r, g, b);
			} 
		}
	}  

	_TIFFfree(scanline);
	TIFFClose(tiffImage);

	if(name)
	  free(name);

	name = strdup(filename);
	NotifyObservers();
	changed = 0;
    
	return 1;
}

int floatimage::SaveTIF(const char *filename)
{
	TIFF *tiffImage = TIFFOpen(filename, "w");

	if (!tiffImage) 
		return 0;

	TIFFSetField(tiffImage, TIFFTAG_IMAGEWIDTH, sizex);
	TIFFSetField(tiffImage, TIFFTAG_IMAGELENGTH, sizey);
	TIFFSetField(tiffImage, TIFFTAG_ROWSPERSTRIP, sizey);
	TIFFSetField(tiffImage, TIFFTAG_SAMPLESPERPIXEL, 3);
	TIFFSetField(tiffImage, TIFFTAG_SAMPLEFORMAT, 3);
	TIFFSetField(tiffImage, TIFFTAG_COMPRESSION, 1);
	TIFFSetField(tiffImage, TIFFTAG_BITSPERSAMPLE, 32);
	TIFFSetField(tiffImage, TIFFTAG_PLANARCONFIG, 1);
	TIFFSetField(tiffImage, TIFFTAG_PHOTOMETRIC, 2);

		
	for (int row = 0; row < sizey; row++)
	{
		TIFFWriteRawStrip(tiffImage, 0, &line[row][0], sizeof(float) * sizex * 3);
	}

	TIFFClose(tiffImage);

	if(name)
	  free(name);

	name = strdup(filename);
	NotifyObservers();
	changed = 0;

	return 1;
}


#endif


int floatimage::LoadBMP_SubImage(int startx, int starty, int width, int height, const char *filename, float image_gamma)
{
  SaveUndo();

	FILE *infile = fopen(filename, "rb");

	if (infile == NULL)
		return 0;

	BITMAPINFOHEADER info;
	if (!ReadHeader(infile, &info))
	{
		fclose(infile);
		return 0;
	}
	
	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int stride = sizeof(RGB) * info.biWidth;			// stride between lines in bytes
	stride = (stride + 3) & (~0x03);					// rounded up to nearest LONG boundary

	InitGamma(image_gamma);

	RGB *buf = new RGB[width];

	fseek(infile, stride * starty, SEEK_CUR);		// read in all the junk before the area we want

	unsigned int ofs = ftell(infile);
	for (int y = 0; y < height; y++)
	{
		fseek(infile, ofs + startx * sizeof(RGB), SEEK_SET);	// jump to start of data

		if (fread(buf, width * sizeof(RGB), 1, infile) != 1)	// read relevant data
		{
			ReleaseImage();
			delete [] buf;
			fclose(infile);
			return 0;
		}

		RGB *cur = buf;											// put data into the floatimage
		fRGB *curline = line[sizey - 1 - y];
		for (int x = 0; x < width; x++)
		{
			curline[x].r = gamma_table[cur->rgbRed];
			curline[x].g = gamma_table[cur->rgbGreen];
			curline[x].b = gamma_table[cur->rgbBlue];
			cur++;
		}

		ofs += stride;											// goto next line
	}

	delete [] buf;
	fclose(infile);

	name = strdup(filename);
	NotifyObservers();
	changed = 0;

	return 1;
}


int floatimage::LoadBMPSize(const char *filename)
{
	FILE *infile = fopen(filename, "rb");

	if (infile == NULL)
		return 0;

	BITMAPINFOHEADER info;
	if (!ReadHeader(infile, &info))
	{
		fclose(infile);
		return 0;
	}

	sizex = info.biWidth;
	sizey = info.biHeight;

	fclose(infile);
	return 1;
}

int floatimage::LoadBMP(const char *filename, float image_gamma)
{
  SaveUndo();

	FILE *infile = fopen(filename, "rb");

	if (infile == NULL)
		return 0;

	BITMAPINFOHEADER info;
	if (!ReadHeader(infile, &info))
	{
		fclose(infile);
		return 0;
	}

	Allocate(info.biWidth, info.biHeight);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int stride = sizeof(RGB) * sizex;			// stride between lines in bytes
	stride = (stride + 3) & (~0x03);			// rounded up to nearest LONG boundary

	InitGamma(image_gamma);

	RGB *buf = (RGB *) new char[stride];

	for (int y = 0; y < sizey; y++)
	{
		if (fread(buf, stride, 1, infile) != 1)
		{
			ReleaseImage();
			delete [] buf;
			fclose(infile);
			return 0;
		}

		RGB *cur = buf;
		fRGB *curline = line[sizey - 1 - y];
		for (int x = 0; x < sizex; x++)
		{
			curline[x].r = gamma_table[cur->rgbRed];
			curline[x].g = gamma_table[cur->rgbGreen];
			curline[x].b = gamma_table[cur->rgbBlue];
			cur++;
		}
	}

	delete buf;

/*	int max = info.biHeight * info.biWidth * 3;
	unsigned char *data = buf();

	for (int x = 0; x < max; x += 3)	// swap red/blue channels
	{
		data[x] ^= data[x+2];
		data[x+2] ^= data[x];
		data[x] ^= data[x+2];
	}
*/
	fclose(infile);

	if(name)
	  free(name);

	name = strdup(filename);
	NotifyObservers();
	changed = 0;

	return 1;
}

int floatimage::SaveBMP(const char *filename, float image_scale, float image_gamma)
{
  unsigned int x, y;

	FILE *outfile = fopen(filename, "wb");

	BITMAPINFOHEADER info;
	info.biSize = sizeof(BITMAPINFOHEADER);
	info.biWidth = sizex;
	info.biHeight = sizey;
	info.biPlanes = 1;
	info.biBitCount = 24;
	info.biCompression = BI_RGB;
	info.biSizeImage = 0;
	info.biXPelsPerMeter = 2834;
	info.biYPelsPerMeter = 2834;
	info.biClrUsed = 0;
	info.biClrImportant = 0;

	if (!WriteHeader(&info, outfile))
	{
		fclose(outfile);
		return 0;
	}

	unsigned char lookup[65536];
	for (x = 0; x < 65536; x++)
	{
		float tval = pow(short2float(x) * image_scale, 1.0 / image_gamma) * 256.0;
		lookup[x] = (tval > 255.0 ? 255 : (unsigned char) tval);
	}


	for (y = sizey - 1; y >= 0; y--)
	{
		float *buf = &line[y][0].r;
		for (x = 0; x < sizex; x++)
		{
			if ((fwrite(&lookup[float2short(buf[2])], 1, 1, outfile) != 1) ||
				(fwrite(&lookup[float2short(buf[1])], 1, 1, outfile) != 1) ||
				(fwrite(&lookup[float2short(buf[0])], 1, 1, outfile) != 1))
			{
				fclose(outfile);
				return 0;
			}

			buf += 3;
		}
	}
	
	fclose(outfile);

	if(name)
	  free(name);

	name = strdup(filename);
	NotifyObservers();

	return 1;
}


int floatimage::SavePPM(const char *filename, float image_scale, float image_gamma)
{
	return 0;
}

int floatimage::SaveJPG(const char *filename, float image_scale, float image_gamma)
{
	return 0;
}

int floatimage::LoadPPM(const char *filename, float image_gamma)
{
	return 0;
}

int floatimage::LoadJPG(const char *filename, float image_gamma)
{
	return 0;
}

void floatimage::SwapData(floatimage &in) 
{
	fRGB **tmp;
	int tmp_x,tmp_y;

	tmp = line; tmp_x = sizex; tmp_y = sizey;
	line = in.line; sizex = in.sizex; sizey = in.sizey;
	in.line = tmp; in.sizex = tmp_x; in.sizey = tmp_y;
}

void floatimage::ResampleFrom(floatimage &in) 
{
	fRGB val;
	float *f = &line[0][0].r;
	float inx,iny;
	int y,x;

	for (y = 0; y < sizey; y++) 
	{
		iny = ((y+0.5f)*in.sizey)/sizey-0.5f;
		for (x = 0; x < sizex; x++) 
		{
			inx = ((x+.5f)*in.sizex)/sizex-0.5f;

			val = in.GetPixelBilinear(inx,iny);

			*f++ = val.r;
			*f++ = val.g;
			*f++ = val.b;
		}
	}
}

void floatimage::gaussian_blur(int variance) {

  SaveUndo();

	int i,count;
	float scalefac;
  
	scalefac = 1.f;
	count = 0;
   
	for (i=0;i<variance;i++) {
		blur1x();
		blur1y();
		count++;

		scalefac*=1.0f/16;

		if (count>5 || i==variance-1) {
			scaleimage(scalefac);
			count = 0; scalefac = 1.f;
		}

	}

	NotifyObservers();
}

void floatimage::fast_gaussian_blur(int variance) {

  SaveUndo();

	int i,count;
	float variance_so_far,rough_target;
	float scalefac;
	int n;
  
	variance_so_far = 0.f;
	scalefac = 1.f;
	count = 0;

	n = 1;

	while (variance/m_boxvariance(n) > 15 ) n++;
	if (variance/m_boxvariance(n) < 5 ) n--;
	if (n==0) n=1;

	printf("%f %d\n",m_boxvariance(n),n);

	if (variance <= 3) rough_target = 0.f;
	else rough_target = variance - (15 - variance/m_boxvariance(n));

	if (rough_target<0.f) rough_target=0.f;

	printf("target %f\n",rough_target);

	while (n>1) {
		while (variance_so_far + m_boxvariance(n) < rough_target) {
			this->box_blur(n);
			printf("subblur %d\n",n);
			variance_so_far += m_boxvariance(n);
		}
		n = n/2;
	}
   
	for (i=(int)(variance_so_far);i<variance;i++) {
		blur1x();
		blur1y();
		count++;
			printf("clean-up blur\n");

		scalefac*=1.0f/16;

		if (count>5 || i==variance-1) {
			scaleimage(scalefac);
			count = 0; scalefac = 1.f;
		}
	}

	NotifyObservers();
}

void floatimage::box_blur(int hw) {

	this->boxblurx(hw);
	this->boxblury(hw);
	this->scaleimage(1.f/((2*hw+1)*(2*hw+1)));

}

void floatimage::boxblurx(int hw) {
	int y,x;
	float *on,*left,*right;
	float *f = &line[0][0].r;
	float *saveout,*out;

	saveout = (float*)malloc(sizex*3*sizeof(float));

	for (y = 0; y < sizey; y++) {

		left = f;
		on = f;
		right = f;
		out = saveout;

		out[0] = hw*left[0];
		out[1] = hw*left[1];
		out[2] = hw*left[2];

		for (x=0;x<=hw;x++) {
			out[0] += *right++;
			out[1] += *right++;
			out[2] += *right++;
		}
		out+=3;
		
		for (x=1;x<=hw;x++) {
			*out++ = *(out-3)+(*right++)-left[0];
			*out++ = *(out-3)+(*right++)-left[1];
			*out++ = *(out-3)+(*right++)-left[2];
		}
		for (x=hw+1;x<sizex-hw-1;x++) {
			*out++ = *(out-3)+(*right++)-(*left++);
			*out++ = *(out-3)+(*right++)-(*left++);
			*out++ = *(out-3)+(*right++)-(*left++);
		}
		for (x=sizex-hw-1;x<sizex;x++) {
			*out++ = *(out-3)+right[0]-(*left++);
			*out++ = *(out-3)+right[1]-(*left++);
			*out++ = *(out-3)+right[2]-(*left++);
		}

		memcpy(f,saveout,3*sizex*sizeof(float));

		f+=3*sizex;
	}

	free(saveout);
}

void floatimage::boxblury(int hw) {
	int y,x;
	float *on,*up,*down,*tmp;
	float *f = &line[0][0].r;
	float *saveout,*out;

	saveout = (float*)malloc(sizey*3*sizeof(float));


	for (x = 0; x < sizex; x++) {

		up = f;
		on = f;
		down = f;
		out = saveout;

		out[0] = hw*up[0];
		out[1] = hw*up[1];
		out[2] = hw*up[2];

		for (y=0;y<=hw;y++) {
			out[0] += *down++;
			out[1] += *down++;
			out[2] += *down++;
			down+=3*(sizex-1);
		}
		out+=3;
		
		for (y=1;y<=hw;y++) {
			*out++ = *(out-3)+(*down++)-up[0];
			*out++ = *(out-3)+(*down++)-up[1];
			*out++ = *(out-3)+(*down++)-up[2];
			down+=3*(sizex-1);
		}
		for (y=hw+1;y<sizey-hw-1;y++) {
			*out++ = *(out-3)+(*down++)-(*up++);
			*out++ = *(out-3)+(*down++)-(*up++);
			*out++ = *(out-3)+(*down++)-(*up++);
			down+=3*(sizex-1); 	up+=3*(sizex-1);
		}
		for (y=sizey-hw-1;y<sizey;y++) {
			*out++ = *(out-3)+down[0]-(*up++);
			*out++ = *(out-3)+down[1]-(*up++);
			*out++ = *(out-3)+down[2]-(*up++);
			up+=3*(sizex-1);
		}
		tmp = f;
		for (y=0;y<sizey;y++) {
			*tmp++ = saveout[3*y+0];
			*tmp++ = saveout[3*y+1];
			*tmp++ = saveout[3*y+2];
			tmp+=3*(sizex-1);
		}

		f+=3;
	}

	free(saveout);

}

void floatimage::blur1x() {
  int  y;
  register float *on,*left,*right;
  register float *out;
  register int  x;
  float *f = &line[0][0].r;
  float *saveout;

  saveout = (float*)malloc(sizex*3*sizeof(float));

  for (y = 0; y < sizey; y++) {
  
    out = saveout;
    on = f;
    left = f;
    right = f+3;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;

    left-=3;
    for (x = 1; x < sizex-1; x++) {
		*out++ = *on + *on + *left++ + *right++; on++;
		*out++ = *on + *on + *left++ + *right++; on++;
		*out++ = *on + *on + *left++ + *right++; on++;
    }
	right-=3;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;

	memcpy(f,saveout,3*sizex*sizeof(float));

	f+=3*sizex;

  }

  	free(saveout);

}

void floatimage::blur1y() {
  int  y;

  register float *on,*up,*down;
  register float *out;
  register int  x;
  float *f = &line[0][0].r;
  float *saveout,*lastout,*tmp;

  saveout = (float*)malloc(sizex*3*sizeof(float));
  lastout = (float*)malloc(sizex*3*sizeof(float));

    out = lastout;
    on = f;
    up = f;
    down = f+3*sizex;
	for (x = 0; x < sizex; x++) {
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
	}

    up-=3*sizex;
	tmp = saveout; saveout = lastout; lastout = tmp;

	for (y = 1; y < sizey-1; y++) {
		out = lastout;
		for (x = 0; x < sizex; x++) {
			*out++ = *on + *on + *up++ + *down++; on++;
			*out++ = *on + *on + *up++ + *down++; on++;
			*out++ = *on + *on + *up++ + *down++; on++;
		}
		memcpy(f,saveout,3*sizex*sizeof(float));
		tmp = saveout; saveout = lastout; lastout = tmp;
		f+=3*sizex;
	}

	down-=3*sizex;
	out = lastout;
	for (x = 0; x < sizex; x++) {
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
	}

	memcpy(f,saveout,3*sizex*sizeof(float));
	memcpy(f+3*sizex,lastout,3*sizex*sizeof(float));

	free(lastout);
	free(saveout);



}

void floatimage::multiplyimage(float cr,float cg,float cb) 
{
  SaveUndo();

	int  y;
	register int  x;
	float *f = &line[0][0].r;


	for (y = 0; y < sizey; y++) 
	{
		for (x = 0; x < sizex; x++) 
		{
			*f++ *= cr;
			*f++ *= cg;
			*f++ *= cb;
		}
	}

	NotifyObservers();
}

void floatimage::scaleimage(float s) 
{
  SaveUndo();

	int  y;
	register int  x;
	float *f = &line[0][0].r;

	for (y = 0; y < sizey; y++) 
	{
		for (x = 0; x < sizex; x++) 
		{
			*f++ *= s;
			*f++ *= s;
			*f++ *= s;
		}
	}
	NotifyObservers();
}

void floatimage::shoulder(float cut) 
{
  SaveUndo();

	float scale,t,invscale,val;
	float *f = &line[0][0].r;
	float *last = &line[sizey-1][sizex-1].b;

	scale=2*(1-cut)/3.14159265f;
	invscale = 1/scale;

	while (f <= last)
	{
		val = *f;
		if (val>cut) 
		{
			t = (val-cut)*invscale;
			*f = scale*(float)atan(t)+cut;
		}
		f++;
	}
	NotifyObservers();
}

void floatimage::vignette(float r0, float r1) 
{
  SaveUndo();

	int  y;
	register int  x;
	float radius;
	float reduce;
	float a,t;
	float *f = &line[0][0].r;

	for (y = 0; y < sizey; y++) 
	{
		for (x = 0; x < sizex; x++) 
		{

			radius = (float)sqrt((x-sizex/2)*(x-sizex/2) + (y-sizey/2)*(y-sizey/2));

			if (radius > r0) 
			{
				if (radius < r1) 
				{
					t = 1.0f - (radius-r0)/(r1-r0);
					a = t * 2 - 1;
					reduce = (float)((0.25 * M_PI + 0.5 * asin(a) + 0.5 * a * sqrt( 1 - a*a ))/(0.5 * M_PI));

				} else 
				{
					reduce =0.f;
				}
				*f++ *= reduce;
				*f++ *= reduce;
				*f++ *= reduce;
			}
			else 
			{
				f+=3;
			}
		}
	}
	NotifyObservers();
}


void floatimage::sepia(float cr,float cg,float cb) 
{
  SaveUndo();

	int  y;
	register int  x;
	float lum;
	COLOR sepiac;
	float *f = &line[0][0].r;

	for (y = 0; y < sizey; y++) 
	{
		for (x = 0; x < sizex; x++) 
		{
			setcolor(sepiac,cr,cg,cb);

			lum = lumcolor(f);
			scalecolor(sepiac,lum);
      
			copycolor(f,sepiac);
			f+=3;
		}
	}
	NotifyObservers();
}

void floatimage::Rotate90CW()
{
  SaveUndo();

	floatimage rotated;
	rotated.Allocate(sizey, sizex);

	if (rotated.line)
	{
		for (int y = 0; y < sizey; y++)
			for (int x = 0; x < sizex; x++)
				rotated.line[x][y] = line[y][x];

		fRGB **temp = rotated.line;
		rotated.line = line;
		line = temp;

		sizey = rotated.sizey;
		sizex = rotated.sizex;
	}
	FlipHorizontal(0);

	NotifyObservers();
}


void floatimage::ToSumTable()
{
	int x, y;

	for (x = 1; x < sizex; x++)
	{
		line[0][x].r += line[0][x-1].r;
		line[0][x].g += line[0][x-1].g;
		line[0][x].b += line[0][x-1].b;
	}

	for (y = 1; y < sizey; y++)
	{
		line[y][0].r += line[y-1][0].r;
		line[y][0].g += line[y-1][0].g;
		line[y][0].b += line[y-1][0].b;
	}

	for (y = 1; y < sizey; y++)
	{
		for (x = 1; x < sizex; x++)
		{
			line[y][x].r += line[y-1][x].r + line[y][x-1].r - line[y-1][x-1].r;
			line[y][x].g += line[y-1][x].g + line[y][x-1].g - line[y-1][x-1].g;
			line[y][x].b += line[y-1][x].b + line[y][x-1].b - line[y-1][x-1].b;
		}
	}
	NotifyObservers();
}

// Discrete Cosine Transform -- see DCT.cpp

void floatimage::DCT()
{
  SaveUndo();
  ::DCT(*this);
  NotifyObservers();
}

void floatimage::IDCT()
{
  SaveUndo();
  ::IDCT(*this);
  NotifyObservers();
}

void floatimage::convolve_naive(floatimage &kernel)
{
  SaveUndo();

	floatimage tmp;

	tmp.Allocate(sizex, sizey);

	float	*rt,*gt,*bt;

	for (int y = 0; y < sizey; y++) 
	{
		for (int x = 0; x < sizex; x++) 
		{
			float	r,g,b;
			float	mr,mg,mb;

			r	=	0;
			g	=	0;
			b	=	0;
			mr	=	0;
			mb	=	0;
			mg	=	0;

			for (int ky = 0; ky < kernel.sizey; ky++) 
			{
				for (int kx = 0; kx < kernel.sizex; kx++)
				{
					int	cx = x + kx - kernel.sizex / 2;
					int	cy = y + ky - kernel.sizey / 2;

					if ((cx < 0) || (cy < 0) || (cx >= sizex) || (cy >= sizey)) 
					{
						mr += kernel.line[ky][kx].r;
						mg += kernel.line[ky][kx].g;
						mb += kernel.line[ky][kx].b;
					}
					else 
					{
						r += kernel.line[ky][kx].r * line[cy][cx].r;
						g += kernel.line[ky][kx].g * line[cy][cx].g;
						b += kernel.line[ky][kx].b * line[cy][cx].b;
					}
				}
			}

			tmp.line[y][x].r = r / (1.0f - mr);
			tmp.line[y][x].g = g / (1.0f - mg);
			tmp.line[y][x].b = b / (1.0f - mb);
		}
	}

	SwapData(tmp);
	NotifyObservers();
}
