#ifndef __FLOATIMAGE_H
#define __FLOATIMAGE_H
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#ifdef FLOATIMAGE_USEIMAGE
#include "image.h"
#include "gamma.h"
#endif

#ifdef FLOATIMAGE_USENOTIFY
#include <glib.h>
#endif


struct fRGB
{
	float r;
	float g;
	float b;

	fRGB()
	{
	}

	fRGB(float r, float g, float b)
	{
		this->r = r;
		this->g = g;
		this->b = b;
	}

	void Set(float r, float g, float b)
	{
		this->r = r;
		this->g = g;
		this->b = b;
	}

        void Scale(float r, float g, float b)
        {
	  this->r *= r;
	  this->g *= g;
	  this->b *= b;
        }

	fRGB operator + (fRGB &other)
	{
		return fRGB(r + other.r, g + other.g, b + other.b);
	}

	fRGB operator - (fRGB &other)
	{
		return fRGB(r - other.r, g - other.g, b - other.b);
	}

	fRGB operator / (float div)
	{
		return fRGB(r / div, g / div, b / div);
	}

	fRGB operator * (float mult)
	{
		return fRGB(r * mult, g * mult, b * mult);
	}

	void operator -= (fRGB &other)
	{
		r -= other.r;
		g -= other.g;
		b -= other.b;
	}

	void operator += (fRGB &other)
	{
		r += other.r;
		g += other.g;
		b += other.b;
	}

	void operator *= (float mult)
	{
		r *= mult;
		g *= mult;
		b *= mult;
	}

	float Intensity()
	{
		return 0.299f * r + 0.587f * g + 0.114f * b;
	}

        void ClampUpper(float val)
        {
	  if(r > val)
	    r = val;
	  if(g > val)
	    g = val;
	  if(b > val)
	    b = val;
        }
};


class floatimage
{
	public:
		fRGB **line;	
		int sizex, sizey;
		char *name;

		int refcount; // when refcount falls to zero,
		              // the floatimage gets nuked
		int changed;  // set to zero every time the image
		              // is saved, set to one when it's altered

		int LoadByExtension(char *filename);
		int SaveByExtension(char *filename);

		int LoadHDR(const char *filename);
		int LoadPFM(const char *filename);
		int LoadRAW(const char *filename, int width, int height, int rgb, int byte_reverse, int header_bytes = 0);
		int LoadRAW(const char *filename);
#ifdef FLOATIMAGE_USETIFF
		int LoadTIF(const char *filename);
#endif
		int LoadBMP(const char *filename, float image_gamma = 2.2f);
		int LoadBMP_SubImage(int startx, 
		  int starty, 
		  int width, 
		  int height, 
		  const char *filename, 
		  float image_gamma = 2.2f);

		int LoadBMPSize(const char *filename);
		int LoadPPM(const char *filename, float image_gamma = 2.2f);
		int LoadJPG(const char *filename, float image_gamma = 2.2f);

		int SaveHDR(const char *filename);
		int SavePFM(const char *filename);
		int SaveRAW(const char *filename);
		int SaveTIF(const char *filename);
		int SaveBMP(const char *filename, float image_scale = 1.0f, float image_gamma = 2.2f);
		int SavePPM(const char *filename, float image_scale = 1.0f, float image_gamma = 2.2f);
		int SaveJPG(const char *filename, float image_scale = 1.0f, float image_gamma = 2.2f);

		void Allocate(int x, int y);
		void ReleaseImage();

		void HalfSize()
		{
		  SaveUndo();

			floatimage temp;
			temp.Allocate(sizex / 2, sizey / 2);

			for (int y = 0; y < temp.sizey; y++)
			{
				int cy = 2 * y;
				for (int x = 0; x < temp.sizex; x++)
				{
					int cx = 2 * x;
					temp.line[y][x] = line[cy][cx] + 
					                  line[cy+1][cx] + 
					                  line[cy][cx+1] + 
					                  line[cy+1][cx+1];

					temp.line[y][x] *= 0.25f;
				}
			}

			SwapData(temp);
			
	      
			NotifyObservers();
		}

		void DoubleSize()
		{
		  SaveUndo();

			floatimage temp;
			temp.Allocate(sizex * 2, sizey * 2);

			for (int y = 0; y < temp.sizey; y++)
			{
				int cy = y / 2;
				for (int x = 0; x < temp.sizex; x++)
				{
					int cx = x / 2;
					temp.line[y][x] = line[cy][cx];
				}
			}

			SwapData(temp);
			NotifyObservers();
		}

		floatimage(char *filename = NULL)
		{

#ifdef FLOATIAMGE_USENOTIFY
		    observers.is_setup = 0;
#endif
		    refcount = 1;
		    changed = 0;

#ifdef FLOATIMAGE_USEUNDO
		    undo_buffer = NULL;
		    undo_x = undo_y = 0;
#endif
   		  line = NULL;
		  sizex = 0;
		  sizey = 0;
		  name = NULL;

			if (filename)
				LoadHDR(filename);
		}

		floatimage(int width, int height)
		  {
#ifdef FLOATIAMGE_USENOTIFY
		    observers.is_setup = 0;
#endif
		    refcount = 1;
		    changed = 0;

#ifdef FLOATIMAGE_USEUNDO
		    undo_buffer = NULL;
		    undo_x = undo_y = 0;
#endif

		    line = NULL;
		    sizex = width;
		    sizey = height;
		    name = NULL;

		    Allocate(sizex, sizey);
		  }

		floatimage(const floatimage &in);

		floatimage& operator=(const floatimage &in);

		~floatimage()
		{
#ifdef FLOATIMAGE_USENOTIFY
		  if(observers.is_setup)
		    g_hook_list_clear(&observers);
#endif

		  free(name);
		  ReleaseImage();

#ifdef FLOATIMAGE_USEUNDO
		  if(undo_buffer)
		    free(undo_buffer);
#endif
		}

		void gaussian_blur(int variance);
		void fast_gaussian_blur(int variance);

		void blur1x();
		void blur1y();
		void box_blur(int hw);
		void boxblurx(int hw);
		void boxblury(int hw);

		void convolve_naive(floatimage &kernel);

		void multiplyimage(float cr,float cg,float cb);
		void scaleimage(float s);
		void shoulder(float cut);
		void vignette(float r0, float r1);
		void sepia(float cr,float cg,float cb);
		void ResampleFrom(floatimage &in);
		void SwapData(floatimage &in);

		void ClipLower(float min, float replace);

		void DCT();
		void IDCT();

		void Clear(float r = 0.0f, float g = 0.0f, float b = 0.0f)
		{
		  SaveUndo();

			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ = r;
				*f++ = g;
				*f++ = b;
			}
			NotifyObservers();
		}

		void copy(floatimage &other)
		{
		  SaveUndo();

			Allocate(other.sizex, other.sizey);

			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ = *o++;
				*f++ = *o++;
				*f++ = *o++;
			}
			NotifyObservers();
		}

		void ScaleChannels(float r, float g, float b)
		  {
		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++)
			line[y][x].Scale(r, g, b);
		  }

#ifdef FLOATIMAGE_USEIMAGE
		void addWithGamma(image &i, float gamma = 1.75f)
		{
		  SaveUndo();

			InitGamma(gamma);

			float *f = &line[0][0].r;
			unsigned char *b = &i.line[0][0].rgbBlue;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				f[2] += gamma_table[*b++];
				f[1] += gamma_table[*b++];
				f[0] += gamma_table[*b++];
				f += 3;
			}
			NotifyObservers();
		}

		void minWithGamma(image &i, float gamma = 1.75f)
		{
		  SaveUndo();

			InitGamma(gamma);

			float *f = &line[0][0].r;
			unsigned char *b = &i.line[0][0].rgbBlue;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				f[2] = min(f[2], gamma_table[*b]);  b++;
				f[1] = min(f[1], gamma_table[*b]);  b++;
				f[0] = min(f[0], gamma_table[*b]);  b++;
				f += 3;
			}
			NotifyObservers();
		}

		void ToImage(image &i, float gamma = 1.75f)
		{
			InitGamma(gamma);

			i.Allocate(sizex, sizey);

			float *f = &line[0][0].r;
			unsigned char *b = &i.line[0][0].rgbBlue;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				b[0] = f[2] * 255.0f; // blue
				b[1] = f[1] * 255.0f; // green
				b[2] = f[0] * 255.0f; // red

				b += 3;
				f += 3;
			}
		}

#endif

		void operator *= (float s)
		{
		  SaveUndo();
			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ *= s;
				*f++ *= s;
				*f++ *= s;
			}
			NotifyObservers();
		}

		void operator -= (floatimage &other)
		{
		  SaveUndo();
		        float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ -= *o++;
				*f++ -= *o++;
				*f++ -= *o++;
			}
			NotifyObservers();
		}

		void operator += (floatimage &other)
		{
		  SaveUndo();
			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ += *o++;
				*f++ += *o++;
				*f++ += *o++;
			}
			NotifyObservers();
		}

		void operator += (fRGB &add)
		{
		  SaveUndo();

		  for(int y = 0; y < sizey; y++)
		    for(int x = 0; x < sizex; x++) {
		      line[y][x] += add;
		    }
		  
		  NotifyObservers();
		}

		void operator /= (floatimage &other)
		{
		  SaveUndo();
			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				if (*o != 0.0f)
					*f /= *o;
				f++; o++;
				if (*o != 0.0f)
					*f /= *o;
				f++; o++;
				if (*o != 0.0f)
					*f /= *o;
				f++; o++;
			}
			NotifyObservers();
		}

		void Add(floatimage &a, floatimage &b)
		{
		  SaveUndo();
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				*d++ = *sa++ + *sb++;		// WOOOOOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!!
				*d++ = *sa++ + *sb++;
				*d++ = *sa++ + *sb++;
			}	
			NotifyObservers();
		}

		void Subtract(floatimage &a, floatimage &b)
		{
		  SaveUndo();
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				*d++ = *sa++ - *sb++;		// WOOOOOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!!
				*d++ = *sa++ - *sb++;
				*d++ = *sa++ - *sb++;
			}	
			NotifyObservers();
		}

		void Multiply(floatimage &a, floatimage &b)
		{
		  SaveUndo();
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				*d++ = *sa++ * *sb++;		// WOOOOOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!!
				*d++ = *sa++ * *sb++;
				*d++ = *sa++ * *sb++;
			}	
			NotifyObservers();
		}

		void Divide(floatimage &a, floatimage &b)		// warning - not safe
		{
		  SaveUndo();
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				if (*sb != 0.0f)
					*d /= *sa / *sb;
				else
					*d = FLT_MAX;
				d++; sa++; sb++;
			}	
			NotifyObservers();
		}

		void absolutevalue()
		{
		  SaveUndo();
			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f = fabs(*f); f++;
				*f = fabs(*f); f++;
				*f = fabs(*f); f++;
			}
			NotifyObservers();
		}

		fRGB maximum(int x0, int y0, int x1, int y1)
		{
			fRGB damax = line[y0][x0];

			for (int y = y0; y <= y1; y++)
			{
				for (int x = x0; x <= x1; x++)
				{
					if (line[y][x].r > damax.r)
						damax.r = line[y][x].r;
					if (line[y][x].g > damax.g)
						damax.g = line[y][x].g;
					if (line[y][x].b > damax.b)
						damax.b = line[y][x].b;
				}
			}

			return damax;
		}


		fRGB minimum(int x0, int y0, int x1, int y1)
		{
			fRGB damin = line[y0][x0];

			for (int y = y0; y <= y1; y++)
			{
				for (int x = x0; x <= x1; x++)
				{
					if (line[y][x].r < damin.r)
						damin.r = line[y][x].r;
					if (line[y][x].g < damin.g)
						damin.g = line[y][x].g;
					if (line[y][x].b < damin.b)
						damin.b = line[y][x].b;
				}
			}

			return damin;
		}

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

		void ScaleToRange(float min, float max)
		{
//			fRGB mincolors = minimum(0, 0, sizex-1, sizey-1);
			fRGB maxcolors = maximum(0, 0, sizex-1, sizey-1);
//			float tmin = min(min(mincolors.r, mincolors.g), mincolors.b);
			float tmax = MAX(MAX(maxcolors.r, maxcolors.g), maxcolors.b);
			
			if(tmax > 1e-3)
			  *this *= (max / tmax);
		}

		void ScaleMeanIntensityTo(float val)
		  {
		    unsigned int npixels = sizex * sizey * 3;
		    float one_over_npixels = 1.0 / (float)npixels;

		    float mean = 1e-7;

		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++) {

			mean += line[y][x].Intensity() * one_over_npixels;

		      }

		    float adjustment = val / mean;

		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++) {
			
			line[y][x] *= adjustment;
			
		      }
		  }

		void ClampUpper(float max)
		  {
		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++) {
			
			line[y][x].ClampUpper(max);
			
		      }
		  }

		fRGB average(int x0, int y0, int x1, int y1)
		{
			int count = 0;
			fRGB dasum(0.0f, 0.0f, 0.0f);

			for (int y = y0; y <= y1; y++)
			{
				for (int x = x0; x <= x1; x++)
				{
					dasum.r += line[y][x].r;
					dasum.g += line[y][x].g;
					dasum.b += line[y][x].b;
					count++;
				}
			}

			dasum.r /= count;
			dasum.g /= count;
			dasum.b /= count;

			return dasum;
		}

		void ToSumTable();

		fRGB SumTable_sum(int x0, int y0, int x1, int y1)
		{
			fRGB dasum;

			dasum = line[y1][x1];

			if (x0 > 0)
				dasum -= line[y1][x0-1];

			if (y0 > 0)
			{
				dasum -= line[y0-1][x1];

				if (x0 > 0)		// both - must subtract overlap
				{
					dasum += line[y0-1][x0-1];
				}
			}

			return dasum;
		}

		fRGB SumTable_average(int x0, int y0, int x1, int y1)
		{
			fRGB dasum;

			dasum = line[y1][x1];

			if (x0 > 0)
				dasum -= line[y1][x0-1];

			if (y0 > 0)
			{
				dasum -= line[y0-1][x1];

				if (x0 > 0)		// both - must subtract overlap
				{
					dasum += line[y0-1][x0-1];
				}
			}

			return dasum / ((x1 - x0 + 1) * (y1 - y0 + 1));
		}

		fRGB GetPixelBilinearSlow(double sx, double sy)
		{
			fRGB result;
			int bx = (int) sx;
			int by = (int) sy;
			int bxo = bx + 1;
			int byo = by + 1;
			sx -= bx;
			sy -= by;

			bx %= sizex;
			by %= sizey;
			bxo %= sizex;
			byo %= sizey;

			double osx = 1.0 - sx;
			double osy = 1.0 - sy;

			result.r = line[by][bx].r * osx * osy +
					   line[by][bxo].r * sx * osy +
					   line[byo][bx].r * osx * sy +
					   line[byo][bxo].r * sx * sy;

			result.g = line[by][bx].g * osx * osy +
					   line[by][bxo].g * sx * osy +
					   line[byo][bx].g * osx * sy +
					   line[byo][bxo].g * sx * sy;

			result.b = line[by][bx].b * osx * osy +
					   line[by][bxo].b * sx * osy +
					   line[byo][bx].b * osx * sy +
					   line[byo][bxo].b * sx * sy;

			return result;
		}

		fRGB GetPixelBilinear(float sx, float sy)
		{
			fRGB result;
			int bx = (int) sx;
			int by = (int) sy;
			int bxo = bx + 1;
			int byo = by + 1;
			float w00,w01,w10,w11;
			sx -= bx;
			sy -= by;

			if (bxo>=sizex) bxo=sizex-1;
			if (byo>=sizey) byo=sizey-1;
			if (sx<0.f) bx = bxo = 0;
			if (sy<0.f) by = byo = 0;

			w11 = sx*sy;

			w00 = 1 - sx - sy + w11;
			w10 = sx - w11;
			w01 = sy - w11;

			result.r = line[by][bx].r * w00 +
					   line[by][bxo].r * w10 +
					   line[byo][bx].r * w01 +
					   line[byo][bxo].r * w11;

			result.g = line[by][bx].g * w00 +
					   line[by][bxo].g * w10 +
					   line[byo][bx].g * w01 +
					   line[byo][bxo].g * w11;

			result.b = line[by][bx].b * w00 +
					   line[by][bxo].b * w10 +
					   line[byo][bx].b * w01 +
					   line[byo][bxo].b * w11;

			return result;
		}

		void SwapLines(int a, int b)
		{
			fRGB temp;

			for (int x = 0; x < sizex; x++)
			{	
				temp = line[b][x];
				line[b][x] = line[a][x];
				line[a][x] = temp;
			}
			
		}

		void SwapColumns(int a, int b)
		{
			fRGB temp;

			for (int y = 0; y < sizey; y++)
			{	
				temp = line[y][b];
				line[y][b] = line[y][a];
				line[y][a] = temp;
			}
			
		}

		void FlipVertical(int notify = 1)
		{
		  if(notify)
		    SaveUndo();

			int top = sizey - 1;
			int bottom = 0;
			while (top > bottom)
			{
				SwapLines(top, bottom);
				top--;
				bottom++;
			}

			if(notify)
			  NotifyObservers();
		}

		
		void FlipHorizontal(int notify = 1)
		{
		  if(notify)
		    SaveUndo();

			int right = sizex - 1;
			int left = 0;
			while (right > left)
			{
				SwapColumns(right, left);
				right--;
				left++;
			}
			
			if(notify)
			  NotifyObservers();
		}

		void Rotate90CW();


		void GetSubImage(int xstart, int ystart, int xwidth, int yheight, floatimage &dest)
		{
			dest.Allocate(xwidth, yheight);
			for (int y = 0; y < yheight; y++)
				for (int x = 0; x < xwidth; x++)
				{
					dest.line[y][x] = line[ystart + y][xstart + x];
				}
		}

		void AddBlitModulate(int xstart, int ystart, floatimage &source, float r, float g, float b)
		{
		  SaveUndo();

			for (int y = 0; y < source.sizey; y++)
			{
				int yval = ystart + y;
				if ((yval < 0) || (yval >= sizey))
					continue;
				for (int x = 0; x < source.sizex; x++)
				{
					int xval = xstart + x;
					if ((xval < 0) || (xval >= sizex))
						continue;

					line[yval][xval].r += source.line[y][x].r * r;
					line[yval][xval].g += source.line[y][x].g * g;
					line[yval][xval].b += source.line[y][x].b * b;
				}
			}
			NotifyObservers();
		}

		void AddBlitScaleModulate(int xstart, int ystart, float scale, floatimage &source, float r, float g, float b)
		{
		  SaveUndo();

			int width  = (int) (source.sizex * scale);
			int height = (int) (source.sizey * scale);

			xstart -= width / 2;
			ystart -= height / 2;

			floatimage blah;

			if (scale < 1.0)
			{
				blah.copy(source);
				source.SwapData(blah);
				source.fast_gaussian_blur((0.7 / scale) * (0.7 / scale));
//				source.box_blur(0.5 / scale);
			}

			for (int y = 0; y < height; y++)
			{
				int yval = ystart + y;
				if ((yval < 0) || (yval >= sizey))
					continue;
				for (int x = 0; x < width; x++)
				{
					int xval = xstart + x;
					if ((xval < 0) || (xval >= sizex))
						continue;

					fRGB temp = source.GetPixelBilinearSlow(x / scale, y / scale);
					line[yval][xval].r += temp.r * r;
					line[yval][xval].g += temp.g * g;
					line[yval][xval].b += temp.b * b;
				}
			}

			if (scale < 1.0)
			{
				source.SwapData(blah);
				blah.ReleaseImage();
			}
			NotifyObservers();
		}

		void AddBlitScaleFastSubPixel(float xstart, float ystart, float scale, floatimage &source)
		{
		  SaveUndo();

			float width = source.sizex * scale;
			float height = source.sizey * scale;

			xstart -= width / 2;
			ystart -= height / 2;

			for (int y = 0; y < height; y++)
			{
				int yval = (int)(ystart + y);
				if ((yval < 0) || (yval >= sizey))
					continue;
				for (int x = 0; x < width; x++)
				{
					int xval = (int)(xstart + x);
					if ((xval < 0) || (xval >= sizex))
						continue;

					fRGB temp = source.line[(int) (y / scale)][(int) (x / scale)];
					line[yval][xval].r += temp.r;
					line[yval][xval].g += temp.g;
					line[yval][xval].b += temp.b;
				}
			}
			NotifyObservers();
		}

		void AddBlitScale(int xstart, int ystart, float scale, floatimage &source)
		{
		  SaveUndo();

			int width = (int)(source.sizex * scale);
			int height = (int)(source.sizey * scale);

			xstart -= width / 2;
			ystart -= height / 2;

			for (int y = 0; y < height; y++)
			{
				int yval = ystart + y;
				if ((yval < 0) || (yval >= sizey))
					continue;
				for (int x = 0; x < width; x++)
				{
					int xval = xstart + x;
					if ((xval < 0) || (xval >= sizex))
						continue;

					fRGB temp = source.line[(int) (y / scale)][(int) (x / scale)];
					line[yval][xval].r += temp.r;
					line[yval][xval].g += temp.g;
					line[yval][xval].b += temp.b;
				}
			}
			NotifyObservers();
		}

		void AddBlitScaleFast(int xstart, int ystart, float scale, floatimage &source)
		{
		  SaveUndo();

			int width = (int)(source.sizex * scale);
			int height = (int)(source.sizey * scale);

			xstart -= width / 2;
			ystart -= height / 2;

			floatimage blah;

			if (scale < 1.0)
			{
				blah.copy(source);
				source.SwapData(blah);
//				source.fast_gaussian_blur((0.7 / scale) * (0.7 / scale));
				source.box_blur(3.0 / scale);
			}

			for (int y = 0; y < height; y++)
			{
				int yval = ystart + y;
				if ((yval < 0) || (yval >= sizey))
					continue;
				for (int x = 0; x < width; x++)
				{
					int xval = xstart + x;
					if ((xval < 0) || (xval >= sizex))
						continue;

					fRGB temp = source.GetPixelBilinearSlow(x / scale, y / scale);
					line[yval][xval].r += temp.r;
					line[yval][xval].g += temp.g;
					line[yval][xval].b += temp.b;
				}
			}

			if (scale < 1.0)
			{
				source.SwapData(blah);
				blah.ReleaseImage();
			}
			NotifyObservers();
		}
		
		void ShiftX(float percent, floatimage &dest)
		{
			while (percent < 0.0)
				percent += 1.0;
			float shift = percent * sizex;
			dest.Allocate(sizex, sizey);
			for (int y = 0; y < sizey; y++)
				for (int x = 0; x < sizex; x++)
				{
					float sx = x + shift;

//					while (sx < 0.0)
//						sx += sizex;
//					while (sx >= sizex)
//						sx -= sizex;

					if (sx < 0.0)
						sx += ((int) (sx / sizex) + 1) * sizex;

					sx -= ((int) (sx / sizex)) * sizex;

					dest.line[y][x] = GetPixelBilinearSlow(sx, y);
				}
		}

		void ShiftX(int offset)
		  {
		    if(offset == 0)
		      return;

		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++) 
			{
			  fRGB temp = line[y][x];

			  line[y][x] = line[y][(x - offset) % sizex];
			  line[y][(x - offset) % sizex] = temp;
			}
		  }

		void ShiftXFrom(floatimage &src, float offset)
		  {
		    Allocate(src.sizex, src.sizey);

		    offset = -offset;

		    if(offset < 0.0)
		      offset += (float)sizex * ((int)(-offset / sizex)+ 5);

		    int ioff = (int) offset;
		    float rem = offset - ioff;

		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++) 
			{
			  fRGB x1, x2;

			  x1 = src.line[y][(x + ioff) % sizex];
			  x2 = src.line[y][(x + ioff + 1) % sizex];
			  
			  x1 *= 1 - rem; x2 *= rem;

			  line[y][x] = x1 + x2;
			}
		  }

		void ShiftXOne()
		{
		  SaveUndo(); // ?

		  int x, y;

			for (y = 0; y < sizey; y++)
			{
				fRGB temp = line[y][0];

				for (x = 0; x < sizex - 1; x++)
				  line[y][x] = line[y][x + 1];
			
				line[y][x] = temp;
			}
			NotifyObservers();
		}

     
#define TWO_PI              6.283185307179586476925286766559
#define RADIANS_PER_DEGREE  0.01745329251994329576923690768
#define RADIANS(x)          RADIANS_PER_DEGREE * (x)

		void AddLightPolar(float theta, float phi, float r, float g, float b, float spread = 10.0)
		  {
		    if(r<1e-4 && g<1e-4 &&b<1e-4)
		      return;

		    float total_energy = 0.0;

		    theta = theta / 360.0;
		    theta = theta - ((int) theta - (theta < 0.0));

		    float center1_x = theta * (float) sizex;
		    float center_y = ((float)sizey * 0.5) * (1.0 - sin(RADIANS(phi)));

		    float center2_x = center1_x + (float) sizex;
		    float center3_x = center1_x - (float) sizex;
		    

		    floatimage light_img;
		    light_img.Allocate(sizex, sizey);

		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++)
			{
			  float dist1   = cos(RADIANS(phi)) * ((float) x - center1_x) * ((float) x - center1_x);
			        dist1  += ((float) y - center_y) * ((float) y - center_y);

			  float dist2   = cos(RADIANS(phi)) * ((float) x - center2_x) * ((float) x - center2_x);
			        dist2  += ((float) y - center_y) * ((float) y - center_y);

			  float dist3   = cos(RADIANS(phi)) *  ((float) x - center3_x) * ((float) x - center3_x);
			        dist3  += ((float) y - center_y) * ((float) y -  center_y);

			  float dist = (dist1 < dist2 ? dist1 : dist2);
				dist = (dist  < dist3  ? dist : dist3);

			  if(dist > spread) {
			    light_img.line[y][x].Set(0.0, 0.0, 0.0);
			    continue;
			  }

			  fRGB mixcol(r, g, b);
			  mixcol *= spread - dist;

			  total_energy += mixcol.Intensity();

			  light_img.line[y][x] = mixcol;
			}

		    // re-normalize
		    fRGB foo(r, g, b);

		    //                             VVVV arbitrary constant
		    light_img *= foo.Intensity() * 20.0 / total_energy;

		    *this += light_img;
		  }


		void LambertIntegrateFrom(floatimage &src)
		  {
		    Allocate(src.sizex, src.sizey);

		    for(int y = 0; y < sizey; y++)
		      for(int x = 0; x < sizex; x++) {

			fRGB sum(0.0, 0.0, 0.0);

			float normal_theta = (float)x / (float)sizex * TWO_PI;
			float normal_phi = (float) y / (float) sizey - 0.5;

			for(int iy = 0; iy < sizey; iy++)
			  for(int ix = 0; ix < sizex; ix++) {

			    ;// XXX DO SOMETHING WILL YA

			  }

		      }
		  }

#ifdef FLOATIMAGE_USENOTIFY
		void RegisterObserver(void *callback, void *user_data)
		  {
		    if(!observers.is_setup) {
		      g_hook_list_init(&observers, sizeof(GHook));
		      observers.is_setup = 1;
		    }

		    GHook *hook = g_hook_alloc(&observers);

		    hook->func = callback;
		    hook->data = user_data;

		    g_hook_prepend(&observers, hook);
		    
		  }

		void UnregisterObserver(void *callback, void *user_data)
		  {
		    GHook *hook = g_hook_find_func_data(&observers, TRUE, callback, user_data);

		    if(hook)
		      g_hook_destroy_link(&observers, hook);
		  }
#endif

		void IncRef()
		  { 
		    refcount++; 

		    // printf("IncRef(); refcount now %d\n", refcount);
		  }

		void DecRef()
		  {
		    refcount--;

		    // printf("DecRef(); refcount now %d\n", refcount);

		    if(refcount <= 0) {
		      // printf("delete this\n");
		      delete this;
		    }
		  }

#ifdef FLOATIMAGE_USEUNDO
		void Undo()
		  {
		    if(!line) return;
		    if(!undo_buffer) return;

		    ReleaseImage();

		    Allocate(undo_x, undo_y);

		    memcpy(line[0], undo_buffer, sizeof(fRGB) * sizex * sizey);

		    NotifyObservers();
		  }
#endif

 private:

#ifdef FLOATIMAGE_USENOTIFY
		GHookList observers;
#endif

#ifdef FLOATIMAGE_USEUNDO
		fRGB *undo_buffer;
		int undo_x, undo_y;
#endif

		void SaveUndo() 
		  {
#ifdef FLOATIAMGE_USEUNDO
		    if(!line)
		      return;

		    if( (!undo_buffer) || (undo_x != sizex) || (undo_y != sizey) ) 
		      {
			if(undo_buffer)
			  free(undo_buffer);
			
			undo_buffer = (fRGB*) malloc(sizeof(fRGB) * sizex * sizey);
			undo_x = sizex;
			undo_y = sizey;
		      }
		    

		    memcpy(undo_buffer, line[0], sizeof(fRGB) * sizex * sizey);
#endif
		  }


		void NotifyObservers()
		  {
#ifdef FLOATIMAGE_USENOTIFY
		    if(observers.is_setup)
		      g_hook_list_invoke(&observers, FALSE);
#endif

		    changed = 1;
		  }

};



#endif
