// IMAGE.CPP
// a class and assocated routines for handling
// 24-bit, interleaved LDR images


#include <stdio.h>
#include <stdlib.h>

#define IN_BMP_C
#include "bmp.h"
#include "image.h"

#define IMAGE_USEHDR

#ifdef IMAGE_USEHDR
#include "rad_color.h"
#include "rad_header.h"
#include "rad_resolu.h"
#include "gamma.h"
#endif

unsigned char image::val[4] = { 0, 0, 1, 1 };


// Floating point format:
//
//			SEEEEEEEEMMMMMMMMMMMMMMMMMMMMMMM
//			|       |       |       |      |
//			0       8       16      24     31
//
//
//		value = (-1)^(S) * (1.MMMMMMMMMMMMMMMMMMMMMMM) * 2^(signed binary EEEEEEEE)
//

inline unsigned short int float2short(float f)
{
  unsigned int bits = * (unsigned int *) &f;
  return (bits >> 15);
}


inline unsigned short int signedfloat2short(float f)
{
  unsigned int bits = * (unsigned int *) &f;
  return (bits >> 16);
}


inline float short2float(unsigned short int s)
{
	unsigned int bits = s << 15;
	return * (float *) &bits;
}

inline float short2signedfloat(unsigned short int s)
{
	unsigned int bits = s << 16;
	return * (float *) &bits;
}

int image::LoadBMP(const char *filename)
{
	FILE *infile = fopen(filename, "rb");

	if (infile == NULL) {
	  printf("LoadBMP: file %s doesn't exist\n", filename);
		return 0;
	}

	if (!ReadHeader(infile, &info))
	{
		fclose(infile);
		return 0;
	}

	Allocate(info.biWidth, info.biHeight);
	if (!line)
	{
		fclose(infile);
		return 0;
	}


	if (fread(line[0], stride * sizey, 1, infile) != 1)
	  {
	    ReleaseImage();
	    fclose(infile);
	    return 0;
	  }
	
#ifdef IMAGE_RGB
	/* swap RGB => BGR (thanks, Bill!) */

	unsigned char *p    = buf();
	unsigned char *end  = buf() + info.biHeight * info.biWidth * 3;
	
	while(p < end) {
	  unsigned char temp = *p;

	  *p = *(p+2);
	  *(p+2) = temp;

	  p += 3;
	  }
#endif

	fclose(infile);

	return 1;

}



int image::SaveBMP(const char *filename)
{
	FILE *outfile = fopen(filename, "wb");
	
	CreateInfoHeader();

	if (!WriteHeader(&info, outfile))
	{
		fclose(outfile);
		return 0;
	}

	// swap RGB => BGR (thanks, Bill!)
	unsigned char *bgr = (unsigned char*) malloc(stride * sizey * 3);
	
	for(unsigned int y = 0; y < sizey; y++)
	  for(unsigned int x = 0; x < sizex; x++)
	    {
	      unsigned char *data = buf() + y*stride + x*3;
	      unsigned char *p = bgr + y*stride + x*3;

	      *p++ = *(data + 2);
	      *p++ = *(data + 1);
	      *p++ = *data;
	    }

	if (fwrite(bgr, stride * sizey, 1, outfile) != 1)
	  {
	        free(bgr);
		fclose(outfile);
		return 0;
	  }

	free(bgr);

	fclose(outfile);
	return 1;
}


int LoadPPM(const char *filename)
{
	return 0;
}


int SavePPM(const char *filename)
{
	return 0;
}


void image::ReleaseImage()
{
	if (line)
	{
		if (line[0])
		{
			free(line[0]);
			line[0] = NULL;
			}
		free(line);
		line = NULL;
	}
	sizex = 0;
	sizey = 0;
}


void image::Clear(unsigned char r, unsigned char g, unsigned char b)
{
  for (unsigned int y = 0; y < sizey; y++)
	for (unsigned int x = 0; x < sizex; x++)
	  {
		  line[y][x].rgbRed = r;
		  line[y][x].rgbBlue = g;
		  line[y][x].rgbGreen = b;
	  }
}

void image::Allocate(int x, int y)
{
	if (line)
	{
		if ((sizex == x) && (sizey == y))
			return;						// already a good size!  Yay!
//		if (sizex * sizey == x * y)		// enough memory, don't reallocate - except for stride issues
//			goto skip_allocation;
		ReleaseImage();
	}

	line = (RGB **) malloc(sizeof(RGB *) * y);

	if (line)
	{
		stride = sizeof(RGB) * x;			// stride between lines in bytes
		// stride = (stride + 3) & (~0x03);	// rounded up to nearest LONG boundary

		line[0] = (RGB *) malloc(stride * y);

skip_allocation:
		if (!line[0])
			ReleaseImage();
		else
		{
			unsigned char *curline = (unsigned char *) line[0];
			for (int cy = 1; cy < y; cy++)
			{
				curline += stride;
				line[cy] = (RGB *) curline;
			}

			sizex = x;
			sizey = y;
			CreateInfoHeader();
		}
	}
}


void image::Decombobulate(image &i1, image &i2, int shift)
{
	for (unsigned int y = 0; y < sizey; y++)
	{
		for (unsigned int x = 0; x < sizex; x++)
		{
			RGB *i1rgb = &i1.line[y][x];
			RGB *i2rgb = &i2.line[y][x];
			int rval = i2rgb->rgbRed - i1rgb->rgbRed;
			int gval = i2rgb->rgbGreen - i1rgb->rgbGreen;
			int bval = i2rgb->rgbBlue - i1rgb->rgbBlue;
/*			line[y][x].rgbRed += (i1rgb->rgbRed >= i2rgb->rgbRed) << shift;
			line[y][x].rgbGreen += (i1rgb->rgbGreen >= i2rgb->rgbGreen) << shift;
			line[y][x].rgbBlue += (i1rgb->rgbBlue >= i2rgb->rgbBlue) << shift;*/
			addshort(x, y, (rval + gval + bval >= 0) << shift);
		}
	}
}




void image::CreateInfoHeader()
{
	info.biSize = sizeof(BITMAPINFOHEADER);
	info.biWidth = sizex;
	info.biHeight = sizey;
	info.biPlanes = 1;
	info.biBitCount = 24;
	info.biCompression = BI_RGB;
	info.biSizeImage = 0;
	info.biXPelsPerMeter = 2834;
	info.biYPelsPerMeter = 2834;
	info.biClrUsed = 0;
	info.biClrImportant = 0;
}




void image::xform(unsigned int *table)
{
	for (unsigned int y = 0; y < sizey; y++)
		for (unsigned int x = 0; x < sizex; x++)
			*((unsigned short int *) &(line[y][x])) = table[*((unsigned short int *) &(line[y][x]))];
}

#ifdef IMAGE_USEAVI
int image::DrawAVIEvenLines(LPBITMAPINFOHEADER lpbi)
{
	if (lpbi == NULL)
		return 0;

	LPBYTE pData = ((LPBYTE) lpbi) + lpbi->biSize;
	unsigned int y;

	if (sizey & 0x01)	// if size is odd
		y = 0;
	else
	{
		y = 1;
		pData += sizex * 4;
	}

	for (; y < sizey; y += 2)
	{
		for (unsigned int x = 0; x < sizex; x++)
		{
			line[y][x].rgbBlue = *pData++;
			line[y][x].rgbGreen = *pData++;
			line[y][x].rgbRed = *pData++;
			pData++;
		}
		pData += sizex * 4;			// skip a line
	}

	return 1;
}


int image::DrawAVIOddLines(LPBITMAPINFOHEADER lpbi)
{
	if (lpbi == NULL)
		return 0;

	LPBYTE pData = ((LPBYTE) lpbi) + lpbi->biSize;
	unsigned int y;

	if (sizey & 0x01)	// if size is odd
	{
		y = 1;
		pData += sizex * 4;
	}
	else
		y = 0;

	for (; y < sizey; y += 2)
	{
		for (unsigned int x = 0; x < sizex; x++)
		{
			line[y][x].rgbBlue = *pData++;
			line[y][x].rgbGreen = *pData++;
			line[y][x].rgbRed = *pData++;
			pData++;
		}
		pData += sizex * 4;			// skip a line
	}

	return 1;
}


int image::MixAVIEvenLines(LPBITMAPINFOHEADER lpbi, float alpha)
{
	if (lpbi == NULL)
		return 0;

	LPBYTE pData = ((LPBYTE) lpbi) + lpbi->biSize;

	if ((sizex != lpbi->biWidth) || (sizey != lpbi->biHeight))
		return 0;

	float omalpha = 1.0f - alpha;
	unsigned int y;

	if (sizey & 0x01)	// if size is odd
		y = 0;
	else
	{
		y = 1;
		pData += sizex * 4;
	}

	for (; y < sizey; y += 2)
	{
		for (unsigned int x = 0; x < sizex; x++)
		{
			line[y][x].rgbBlue = line[y][x].rgbBlue * (omalpha) + (*pData++) * alpha;
			line[y][x].rgbGreen = line[y][x].rgbGreen * (omalpha) + (*pData++) * alpha;
			line[y][x].rgbRed = line[y][x].rgbRed * (omalpha) + (*pData++) * alpha;
			pData++;
		}
		pData += sizex * 4;			// skip a line
	}

	return 1;
}

int image::MixAVIOddLines(LPBITMAPINFOHEADER lpbi, float alpha)
{
	if (lpbi == NULL)
		return 0;

	LPBYTE pData = ((LPBYTE) lpbi) + lpbi->biSize;

	if ((sizex != lpbi->biWidth) || (sizey != lpbi->biHeight))
		return 0;

	float omalpha = 1.0f - alpha;
	unsigned int y;

	if (sizey & 0x01)	// if size is odd
	{
		y = 1;
		pData += sizex * 4;
	}
	else
		y = 0;

	for (; y < sizey; y += 2)
	{
		for (unsigned int x = 0; x < sizex; x++)
		{
			line[y][x].rgbBlue = line[y][x].rgbBlue * (omalpha) + (*pData++) * alpha;
			line[y][x].rgbGreen = line[y][x].rgbGreen * (omalpha) + (*pData++) * alpha;
			line[y][x].rgbRed = line[y][x].rgbRed * (omalpha) + (*pData++) * alpha;
			pData++;
		}
		pData += sizex * 4;			// skip a line
	}

	return 1;
}


//
//	DV interlaced subframes are encoded as
//
//                Subframe #    0 1 2 3 4 5 6 7 ...
//
//			Subframe Parity:	O E O E O E O E ...
//                              \-/ \-/ \-/ \-/
//                   Frame #     0   1   2   3  ...
//
int image::GetAVIFrameBilinear(IGetFrame FAR *pgf, float lFrame)
{		
	LPBITMAPINFOHEADER lpbi;
	float subframe = lFrame * 2.0f;
	float interframe = (subframe - (int) subframe) / 2.0f;

	if (((int) subframe) & 0x01)	// if subframe is odd then the situation is:  |O E.|O E |
	{
		GetAVIFrame(pgf, (int) lFrame);
		lpbi = (LPBITMAPINFOHEADER) AVIStreamGetFrame(pgf, (int) lFrame + 1);
		if (lpbi == NULL)
			return 0;

		MixAVIEvenLines(lpbi, interframe);
		MixAVIOddLines(lpbi, 0.5f + interframe);
	}
	else							// otherwise the situation is: E|O.E |O
	{
		lpbi = (LPBITMAPINFOHEADER) AVIStreamGetFrame(pgf, (int) lFrame - 1);
		if (lpbi == NULL)
			return 0;

		Allocate(lpbi->biWidth, lpbi->biHeight);
		CreateInfoHeader();

		DrawAVIEvenLines(lpbi);

		lpbi = (LPBITMAPINFOHEADER) AVIStreamGetFrame(pgf, (int) lFrame);
		if (lpbi == NULL)
			return 0;

		DrawAVIOddLines(lpbi);
		MixAVIEvenLines(lpbi, 0.5f + interframe);

		lpbi = (LPBITMAPINFOHEADER) AVIStreamGetFrame(pgf, (int) lFrame + 1);
		if (lpbi == NULL)
			return 0;

		MixAVIOddLines(lpbi, interframe);
	}
		
	return 1;
}

#endif



void image::Rotate90CW()
{
	image rotated;
	rotated.Allocate(sizey, sizex);

	if (rotated.line)
	{
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
				rotated.line[x][y] = line[y][x];

		RGB **temp = rotated.line;
		rotated.line = line;
		line = temp;

		sizey = rotated.sizey;
		sizex = rotated.sizex;
	}
	FlipHorizontal();
}


/*
void image::DrawGoodCorrespondances(correspondance **c, int width, int height)
{
	Allocate(width, height);
	for (unsigned int y = 0; y < sizey; y++)
		for (unsigned int x = 0; x < sizex; x++)
		{
//			if (c[y][x].count > 1)
			{
			  line[y][x].rgbBlue = (c[y][x].stddev <= 16.0 ? c[y][x].stddev * 16 : 255);
			  line[y][x].rgbGreen = (c[y][x].x / 8);
			  line[y][x].rgbRed = (c[y][x].y / 8);
			}
		}
}


void image::DrawGoodCorrespondancesMap(correspondance **c, int width, int height, image &img)
{
	Allocate(width, height);
	for (unsigned int y = 0; y < sizey; y++)
		for (unsigned int x = 0; x < sizex; x++)
		{
		  int cx = c[y][x].x;
		  int cy = c[y][x].y;
		  if ((cx < img.sizex) && (cx >= 0) &&
			  (cy < img.sizey) && (cy >= 0))
		  {
		    line[y][x].rgbBlue = img.line[cy][cx].rgbBlue;
		    line[y][x].rgbGreen = img.line[cy][cx].rgbGreen;
		    line[y][x].rgbRed = img.line[cy][cx].rgbRed;
		  }
		}
}


void image::DrawOneToOneCorrespondances(int width, int height)
{
	Allocate(width, height);
	for (unsigned int y = 0; y < sizey; y++)
		for (unsigned int x = 0; x < sizex; x++)
		{
		  line[y][x].rgbBlue = 0;
		  line[y][x].rgbGreen = (x / 8);
		  line[y][x].rgbRed = (y / 8);
		}
}

  */



#ifdef IMAGE_USEHDR

int image::SaveHDR(const char *filename, float gamma)
{
	FILE *outfile = fopen(filename, "wb");

    fprintf(outfile, "#?RADIANCE");
	fputc(10, outfile);
    fprintf(outfile, "# %s", "Made with 100% pure p1_resample");
	fputc(10, outfile);
    fprintf(outfile, "FORMAT=32-bit_rle_rgbe");
	fputc(10, outfile);
    fprintf(outfile, "EXPOSURE=%25.13f", 1.0);
	fputc(10, outfile);
	fputc(10, outfile);
    fprintf(outfile, "-Y %d +X %d", sizey, sizex);
	fputc(10, outfile);
	
	COLR *tempbuf = new COLR[sizex];

	InitGamma(gamma);

	for (unsigned int y = 0; y < sizey; y++)
	{
		for (unsigned int x = 0; x < sizex; x++)
		{
			RGB *cur = &line[y][x];
			setcolr(tempbuf[x], gamma_table[cur->rgbRed], gamma_table[cur->rgbGreen], gamma_table[cur->rgbBlue]);
		}

		if (fwritecolrs(tempbuf, sizex, outfile) < 0)
		{
			fclose(outfile);
			return 0;
		}
	}

	delete [] tempbuf;

	fclose(outfile);

	return 1;
}


int image::SetHDR(DWORD *buf, int width, int height, float scale, float gamma)
{
	Allocate(width, height);
	if (!line)
		return 0;

	unsigned char lookup[256][256];


	float eval = ldexp(scale, 1 - (COLXS + 8));

	for (int m = 0; m < 256; m++)
		lookup[0][m] = 0;

	for (int e = 1; e < 256; e++)
	{
		float val = 0.5f * eval;
		for (int m = 0; m < 256; m++)
		{
			float tval = pow(val, 1.0 / gamma) * 256.0;
			lookup[e][m] = (tval > 255.0 ? 255 : (unsigned char) tval);
			val += eval;
		}
		eval *= 2.0f;
	}

	/*
	for (int m = 0; m < 256; m++)		// mantissa
	{
		lookup[m][0] = 0.0f;	// exponent of 0 means 0

		float eval = ldexp(scale , (int) 1 - (COLXS + 8));

		for (int e = 1; e < 256; e++)	// exponent
		{			
			float val = (m + 0.5f) * eval;
			eval *= 2.0f;
			val = pow(val, 1.0 / gamma) * 255.0;							// do gamma correction
			lookup[m][e] = (val > 255.0 ? 255 : (unsigned char) val);		// clamp positive at 255 (shouldn't go negative I think)
		}
	}
	*/

	unsigned char *c = (unsigned char *) buf;

	for (unsigned int y = 0; y < sizey; y++)
	{
		for (unsigned int x = 0; x < sizex; x++)
		{
 			unsigned char e = c[3];

#ifndef IMAGE_RGB
			line[sizey-1-y][x].rgbRed = lookup[e][*c++];		// get red
			line[sizey-1-y][x].rgbGreen = lookup[e][*c++];		// get green
			line[sizey-1-y][x].rgbBlue = lookup[e][*c++];		// get blue
			c++;	// skip exponent
#else
			line[sizey-1-y][x].rgbBlue = lookup[e][*c++];		// get red
			line[sizey-1-y][x].rgbGreen = lookup[e][*c++];		// get green
			line[sizey-1-y][x].rgbRed = lookup[e][*c++];		// get blue
			c++;	// skip exponent
#endif
		}
	}

	CreateInfoHeader();

	return 1;
}


int image::SetHDR(float *buf, int width, int height, float scale, float gamma)
{
  unsigned int x, y;

	Allocate(width, height);
	if (!line)
		return 0;

	unsigned char lookup[65536];

	for (x = 0; x < 65536; x++)
	{
		float tval = pow(short2float(x) * scale, 1.0 / gamma) * 256.0;
		lookup[x] = (tval > 255.0 ? 255 : (unsigned char) tval);
	}

	for (y = 0; y < sizey; y++)
	{
		for (x = 0; x < sizex; x++)
		{
#ifndef IMAGE_RGB
			line[sizey-1-y][x].rgbRed = lookup[float2short(*buf++)];		// get red
			line[sizey-1-y][x].rgbGreen = lookup[float2short(*buf++)];		// get green
			line[sizey-1-y][x].rgbBlue = lookup[float2short(*buf++)];		// get blue
#else
			line[sizey-1-y][x].rgbBlue = lookup[float2short(*buf++)];		// get red
			line[sizey-1-y][x].rgbGreen = lookup[float2short(*buf++)];		// get green
			line[sizey-1-y][x].rgbRed = lookup[float2short(*buf++)];		// get blue
#endif
		}
	}

	CreateInfoHeader();

	return 1;
}


int image::SetHDR_negColor(float * buf, int width, int height, float scale, float gamma, unsigned long c)
{
  unsigned int x, y;

	Allocate(width, height);
	if (!line)
		return 0;

	unsigned char nr = ((unsigned char *) &c)[0];
	unsigned char ng = ((unsigned char *) &c)[1];
	unsigned char nb = ((unsigned char *) &c)[2];
	unsigned char rlookup[65536];
	unsigned char glookup[65536];
	unsigned char blookup[65536];

	for (x = 0; x < 32767; x++)
	{
		float tval = pow(short2signedfloat(x) * scale, 1.0 / gamma) * 256.0f;
		rlookup[x] = glookup[x] = blookup[x] = (tval > 255.0f ? 255 : (unsigned char) tval);
		rlookup[32768+x] = nr;
		glookup[32768+x] = ng;
		blookup[32768+x] = nb;
	}

	for (y = 0; y < sizey; y++)
	{
		for (x = 0; x < sizex; x++)
		{
			line[sizey-1-y][x].rgbRed = rlookup[signedfloat2short(*buf++)];		// get red
			line[sizey-1-y][x].rgbGreen = glookup[signedfloat2short(*buf++)];		// get green
			line[sizey-1-y][x].rgbBlue = blookup[signedfloat2short(*buf++)];		// get blue
		}
	}

	CreateInfoHeader();

	return 1;
}

int image::SetHDR_negOfs(float * buf, int width, int height, float scale, float gamma)
{
  unsigned int x, y;

	Allocate(width, height);
	if (!line)
		return 0;

	unsigned char lookup[65536];

	for (x = 0; x < 65536; x++)
	{
		float val = short2signedfloat(x) * scale + 0.5f;
		float tval = pow(fabs(val), 1.0 / gamma) * (val < 0.0f ? -256.0f : 256.0f);
		lookup[x] = (tval > 255.0f ? 255 : (tval < 0.0f ? 0 : (unsigned char) tval));
	}

	for (y = 0; y < sizey; y++)
	{
		for (x = 0; x < sizex; x++)
		{
			line[sizey-1-y][x].rgbRed = lookup[signedfloat2short(*buf++)];	// get red
			line[sizey-1-y][x].rgbGreen = lookup[signedfloat2short(*buf++)];		// get green
			line[sizey-1-y][x].rgbBlue = lookup[signedfloat2short(*buf++)];		// get blue
		}
	}

	CreateInfoHeader();

	return 1;
}

int image::SetHDR_negNeg(float * buf, int width, int height, float scale, float gamma)
{
  unsigned int x, y;

	Allocate(width, height);
	if (!line)
		return 0;

	unsigned char lookup[65536];

	for (x = 0; x < 65536; x++)
	{
		float tval = pow(short2float(x) * scale, 1.0 / gamma) * 256.0;
		lookup[x] = (tval > 255.0 ? 255 : (unsigned char) tval);
	}

	float r, g, b, t;

	for (y = 0; y < sizey; y++)
	{
		for (x = 0; x < sizex; x++)
		{
			t = *buf++;
			if (t < 0.0f)
			{
				t *= 0.5f;
				g = -t;
				b = -t;
				r = 0.0f;
			}
			else
			{
				r = t;
				g = 0.0f;
				b = 0.0f;
			}

			t = *buf++;
			if (t < 0.0)
			{
				t *= 0.5f;
				r -= t;
				b -= t;
			}
			else
				g += t;

			t = *buf++;
			if (t < 0.0)
			{
				t *= 0.5f;
				r -= t;
				g -= t;
			}
			else
				b += t;

			line[sizey-1-y][x].rgbRed = lookup[float2short(r)];		// get red
			line[sizey-1-y][x].rgbGreen = lookup[float2short(g)];		// get green
			line[sizey-1-y][x].rgbBlue = lookup[float2short(b)];		// get blue
		}
	}

	CreateInfoHeader();

	return 1;
}


int image::SetHDR_thru_LUT(float * buf, int width, int height, float t[256], float scale = 1.0)
{ 
	float *p; 
	unsigned char *q; 

	Allocate(width, height);
	if (!line)
	  return 0;

	p = buf;
	q = &(line[0]->rgbRed);

	while(height--)
	{
	  for( unsigned int i = 0; i < width*3; i++ )
	    {
		unsigned char j = 0; 
		float v = scale * *p++; 
 
		if(v > t[128])  
		  j |= 128;  
		if(v > t[(64 | j)])  
		  j |= 64;  
		if(v > t[(32 | j)])  
		  j |= 32;  
		if(v > t[(16 | j)])  
		  j |= 16;  
		if(v > t[(8 | j)])  
		  j |= 8;  
		if(v > t[(4 | j)])  
		  j |= 4;  
		if(v > t[(2 | j)])  
		  j |= 2;  
		if(v > t[(1 | j)])  
		  j |= 1;  
	 
	/*	 The above is a binary search, equivalent to: 
				while((v > t[j]) && (j < 255)) 
					j++;	* 
			(we're finding the greatest value < v in the lookup table */ 
 
		*q++ = j; 
	    }
	 
	}

	CreateInfoHeader();
}


#endif
