#ifndef __IMAGE_H
#define __IMAGE_H

#include "gamma.h"
#include "bmp.h"
#include "math.h"

class correspondance;

// compresses two 12 bit numbers into a 24 bit number
inline unsigned int compress(unsigned int h, unsigned int v)	
{
	return (h + (v << 12));
}

inline void decompress(unsigned int val, unsigned int &h, unsigned int &v)
{
	h = (val & 0x000fff);
	v = (val & 0xfff000) >> 12;
}

class image
{
  public:

	RGB **line;
	unsigned int sizex, sizey, stride;

	BITMAPINFOHEADER info;
	char* name;

	static unsigned char val[4];

	int LoadBMP(const char *filename);
	int SaveBMP(const char *filename);
	int LoadPPM(const char *filename);
	int SavePPM(const char *filename);

	int SaveHDR(const char *filename, float gamma = 1.75);
	int SetHDR(DWORD *buf, int width, int height, float scale = 1.0, float gamma = 2.2);
	int SetHDR(float *buf, int width, int height, float scale = 1.0, float gamma = 2.2);
	int SetHDR_negColor(float * buf, int width, int height, float scale = 1.0, float gamma = 2.2, unsigned long c = 0x00000000);
	int SetHDR_negOfs(float * buf, int width, int height, float scale = 1.0, float gamma = 2.2);
	int SetHDR_negNeg(float * buf, int width, int height, float scale = 1.0, float gamma = 2.2);

	int SetHDR_thru_LUT(float * buf, int width, int height, float t[256], float scale = 1.0);

	void Allocate(int x, int y);
	void ReleaseImage();
	void Clear(unsigned char r, unsigned char g, unsigned char b);

	image(char *filename = NULL)
	{
		line = NULL;
		sizex = 0;
		sizey = 0;
		name = "Unnamed";

		if (filename)
			LoadBMP(filename);
	}

	~image()
	{
		ReleaseImage();
	}

	unsigned char *buf()
	{
		return (unsigned char *) line[0];
	}

	unsigned short int getshort(int x, int y)
	{
		return *((unsigned short int *) &(line[y][x]));
	}

	unsigned int getint(int x, int y)
	{
		return (*((unsigned int *) &(line[y][x]))) & 0x00ffffff;
	}

	void addshort(int x, int y, unsigned short int value)
	{
		*((unsigned short int *) &(line[y][x])) += value;
	}


	void setshort(int x, int y, unsigned short int value)
	{
		*((unsigned short int *) &(line[y][x])) = value;
	}

	void xform(unsigned int *table);

	void operator -= (image &other)
	{
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				register int val = ((int) line[y][x].rgbRed) - other.line[y][x].rgbRed;
				line[y][x].rgbRed = (val < 0 ? 0 : val);

				val = ((int) line[y][x].rgbGreen) - other.line[y][x].rgbGreen;
				line[y][x].rgbGreen = (val < 0 ? 0 : val);

				val = ((int) line[y][x].rgbBlue) - other.line[y][x].rgbBlue;
				line[y][x].rgbBlue = (val < 0 ? 0 : val);
			}
	}

	void operator += (image &other)
	{
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				register int val = ((int) line[y][x].rgbRed) - other.line[y][x].rgbRed;
				line[y][x].rgbRed = (val < 0 ? 0 : val);

				val = ((int) line[y][x].rgbGreen) - other.line[y][x].rgbGreen;
				line[y][x].rgbGreen = (val < 0 ? 0 : val);

				val = ((int) line[y][x].rgbBlue) - other.line[y][x].rgbBlue;
				line[y][x].rgbBlue = (val < 0 ? 0 : val);
			}
	}

	void GreaterThan(unsigned char threshold)
	{
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				line[y][x].rgbRed = (line[y][x].rgbRed < threshold);
				line[y][x].rgbGreen = (line[y][x].rgbGreen < threshold);
				line[y][x].rgbBlue = (line[y][x].rgbBlue < threshold);
			}
	}

	void operator <<= (int shift)
	{
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				line[y][x].rgbRed = (line[y][x].rgbRed << shift);
				line[y][x].rgbGreen = (line[y][x].rgbGreen << shift);
				line[y][x].rgbBlue = (line[y][x].rgbBlue << shift);
			}
	}

	void ToShorts()
	{
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				line[y][x].rgbRed = val[line[y][x].rgbRed | line[y][x].rgbGreen | line[y][x].rgbBlue];
				line[y][x].rgbGreen = 0;
				line[y][x].rgbBlue = 0;
			}
	}


	float *ToFloatRGBArrayViaLookup(float *lookup, float *buffer = NULL)
	{
		if (buffer == NULL)
			buffer = new float[sizex * sizey * 3];
		
		float *cur = buffer;
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				*cur++ = lookup[line[y][x].rgbRed];
				*cur++ = lookup[line[y][x].rgbGreen];
				*cur++ = lookup[line[y][x].rgbBlue];
			}

		return buffer;
	}


	float *ToFloatBGRArrayViaLookup(float *lookup, float *buffer = NULL)
	{
		if (buffer == NULL)
			buffer = new float[sizex * sizey * 3];
		
		float *cur = buffer;
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				*cur++ = lookup[line[y][x].rgbRed];
				*cur++ = lookup[line[y][x].rgbGreen];
				*cur++ = lookup[line[y][x].rgbBlue];
			}

		return buffer;
	}

	
	float *ToFloatArray(float gamma, float *buffer = NULL)
	{
		InitGamma(gamma);

		return ToFloatRGBArrayViaLookup(gamma_table, buffer);
	}


	void CompressHV(image &vert)
	{
		int val;
		unsigned char *t = (unsigned char *) &val; // gotta love C

		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				val = compress(getshort(x, y), vert.getshort(x,y));
				line[y][x].rgbBlue = t[0];
				line[y][x].rgbGreen = t[1];
				line[y][x].rgbRed = t[2];
			}
	}

	void operator &= (image &mask)
	{
		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				line[y][x].rgbRed &= mask.line[y][x].rgbRed;
				line[y][x].rgbGreen &= mask.line[y][x].rgbGreen;
				line[y][x].rgbBlue &= mask.line[y][x].rgbBlue;
			}
	}

	void Decombobulate(image &i1, image &i2, int shift);
	void CreateInfoHeader();

	void DrawGoodCorrespondances(correspondance **c, int width, int height);
	void DrawOneToOneCorrespondances(int width, int height);
	void DrawGoodCorrespondancesMap(correspondance **c, int width, int height, image &img);

#ifdef IMAGE_USEAVI

	int GetAVIFrame(IGetFrame FAR * pgf, int lFrame)
	{
		LPBITMAPINFOHEADER lpbi;
		lpbi = (LPBITMAPINFOHEADER) AVIStreamGetFrame(pgf, lFrame);
		
		if (lpbi == NULL)
			return 0;

		Allocate(lpbi->biWidth, lpbi->biHeight);

		LPBYTE pData = ((LPBYTE) lpbi) + lpbi->biSize;

		for (unsigned int y = 0; y < sizey; y++)
			for (unsigned int x = 0; x < sizex; x++)
			{
				line[y][x].rgbBlue = *pData++;		// could do this in one dword copy, unaligned though
				line[y][x].rgbGreen = *pData++;
				line[y][x].rgbRed = *pData++;
				pData++;							// eat up reserved byte
			}

//		memcpy(line[0], pData, 3 * sizex * sizey);
//		line[0] = (RGB *) pData;

		return 1;
	}

	int DrawAVIEvenLines(LPBITMAPINFOHEADER lpbi);
	int DrawAVIOddLines(LPBITMAPINFOHEADER lpbi);
	int MixAVIEvenLines(LPBITMAPINFOHEADER lpbi, float alpha);
	int MixAVIOddLines(LPBITMAPINFOHEADER lpbi, float alpha);

	int GetAVIFrameBilinear(IGetFrame FAR *pgf, float lFrame);
#endif

	double sum()
	{
		double sum = 0.0;
		int length = sizex * sizey * 3;

		for (int i = 0; i < length; i++)
		{
			sum += buf()[i];
		}

		return sum;
	}


	void SwapLines(int a, int b)
	{
		RGB temp;

		for (unsigned int x = 0; x < sizex; x++)
		{	
			temp = line[b][x];
			line[b][x] = line[a][x];
			line[a][x] = temp;
		}
	}

	void SwapColumns(int a, int b)
	{
		RGB temp;

		for (unsigned int y = 0; y < sizey; y++)
		{	
			temp = line[y][b];
			line[y][b] = line[y][a];
			line[y][a] = temp;
		}
	}

	void FlipVertical()
	{
		int top = sizey - 1;
		int bottom = 0;
		while (top > bottom)
		{
			SwapLines(top, bottom);
			top--;
			bottom++;
		}
	}

	void FlipHorizontal()
	{
		int right = sizex - 1;
		int left = 0;
		while (right > left)
		{
			SwapColumns(right, left);
			right--;
			left++;
		}
	}

	void Rotate90CW();
};




#endif
