#ifndef _XP_THREADS_H
#define _XP_THREADS_H

// Threads across pthreads and Win32
// Dan Maas <dmaas@dcine.com>

#ifndef _WIN32 // PTHREADS Specific //////////////////////////
 
#include <pthread.h> 

//// THREADS

typedef pthread_t thread_t;

#define THREAD_PROTOTYPE(name, data)  void* (name)(void* data)

static inline void 
thread_start(thread_t *t, void *func, void *arg) {
  pthread_t foo;

  if(t)
    pthread_create(t, NULL, (func), (arg));
  else
    pthread_create(&foo, NULL, (func), (arg));
}

static inline void
thread_this(thread_t *t) {
  *t = pthread_self();
}

static inline void
thread_kill(thread_t *t) {
  pthread_cancel(*t); 
}

#define THREAD_PRIORITY(prio) \
struct sched_param sp; \
  pthread_t me; \
  int policy; \
  me = pthread_self();\
  pthread_getschedparam(me, &policy, &sp);\
  sp.sched_priority = prio; \
  pthread_setschedparam(me, policy, &sp);

static inline void thread_priority_normal() {THREAD_PRIORITY(1)}
static inline void thread_priority_low()    {THREAD_PRIORITY(20)}
static inline void thread_priority_high()   {THREAD_PRIORITY(-20)}


// MUTEXII

typedef pthread_mutex_t mutex_t;

static inline int mutex_init(mutex_t *mtx)
{ return pthread_mutex_init(mtx, NULL); }

static inline int mutex_lock(mutex_t *mtx)
{ return pthread_mutex_lock(mtx); }

static inline int mutex_unlock(mutex_t *mtx)
{ return pthread_mutex_unlock(mtx); }

static inline int mutex_destroy(mutex_t *mtx)
{ return pthread_mutex_destroy(mtx); }

// CONDITION VARIABLES

typedef pthread_cond_t cond_t;

static inline int cond_init(cond_t *cv)
{ return pthread_cond_init(cv, NULL); }

static inline int cond_wait(cond_t *cv, mutex_t *mtx)
{ return pthread_cond_wait(cv, mtx); }

static inline int cond_signal(cond_t *cv)
{ return pthread_cond_signal(cv); }

static inline int cond_broadcast(cond_t *cv)
{ return pthread_cond_broadcast(cv); }

static inline int cond_destroy(cond_t *cv)
{ return pthread_cond_destroy(cv); }

#endif // PTHREADS


#ifdef _WIN32 // WIN32 Specific  //////////////////////////
 
#define WIN32_LEAN_AND_MEAN 
#ifdef MSC_VER
extern "C" DWORD WINAPI SignalObjectAndWait(
  HANDLE hObjectToSignal,  // handle to object to signal
  HANDLE hObjectToWaitOn,  // handle to object to wait for
  DWORD dwMilliseconds,    // time-out interval in milliseconds
  BOOL bAlertable          // alertable flag
);
#else
#include <windows.h> 
#include <process.h> 
#include <fcntl.h>
#include <stdio.h>
#endif


//// THREADS

typedef struct { 
  HANDLE handle; 
  DWORD threadId; }
thread_t;

#define THREAD_PROTOTYPE(name, data)	  DWORD WINAPI name(LPVOID data) 

static __inline void
thread_start(thread_t *t, void *func, void *arg) {
  DWORD foo;

  if(t)
    t->handle = CreateThread(NULL, 0, (unsigned long (__stdcall *)(void *)) func, arg, 0, &t->threadId);
  else
    CreateThread(NULL, 0, (unsigned long (__stdcall *)(void *)) func, arg, 0, &foo);
}

static inline void
thread_this(thread_t *t) {
  t->handle = GetCurrentThread();
  t->threadId = GetCurrentThreadId();
}

static inline void
thread_kill(thread_t *t) { TerminateThread(t->handle, 0); }

static inline void
thread_priority_normal() { SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL); }

static inline void
thread_priority_low() { SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_LOWEST); }

static inline void
thread_priority_high() { SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST); }

// MUTEXII

typedef HANDLE mutex_t;

static inline int mutex_init(mutex_t *mtx)
{ *mtx = CreateMutex(NULL, FALSE, NULL); return 1; }

static inline int mutex_lock(mutex_t *mtx)
{ return WaitForSingleObject(*mtx, INFINITE) == WAIT_OBJECT_0; }

static inline int mutex_unlock(mutex_t *mtx)
{ return ReleaseMutex(*mtx); }

static inline int mutex_destroy(mutex_t *mtx)
{ return CloseHandle(*mtx); }

// CONDITION VARIABLES

// NOTE: Due to lack of a SignalObjectAndWaitMultiple() function
// in NT, we can implement EITHER wake-one OR wake-all condition
// variables. It IS possible to do both, but only with mucho overhead.
// 


typedef HANDLE cond_t;

static inline int cond_init(cond_t *cv)
{ *cv = CreateEvent(NULL, TRUE, FALSE, NULL);return 1; }
//                        ^^^^^ TRUE for wake-all, FALSE for wake-one

// Note: SignalObjectAndWait is not available on Win98. 
//      Get a real operating system.

inline int cond_wait(cond_t *cv, mutex_t *mtx)
{ 
  if( (SignalObjectAndWait((HANDLE) *mtx, (HANDLE) *cv, (DWORD) INFINITE, (BOOL) FALSE))
      == WAIT_OBJECT_0) {
    return mutex_lock(mtx);
  }
  return -1; //SOAW failed
}

static inline int cond_signal(cond_t *cv)
{ 
  // to implement wake-one, change the CreateEvent call
  // in cond_init (see comment above) and make this
  // function a PulseEvent();
  fprintf(stderr, "cond_signal not implemented on Win32!\n");
  exit(-1);
  return -1;
}

static inline int cond_broadcast(cond_t *cv)
{ return PulseEvent(*cv); }
              
static inline int cond_destroy(cond_t *cv)
{ return CloseHandle(*cv); }
   
#endif // WIN32

#endif // _XP_THREADS_H
